# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
topology platforms module entry point.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division



#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Base platform engine module for topology.

This module defines the functionality that a topology platform plugin must
provide to be able to create a network using specific hardware or software.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from abc import ABCMeta, abstractmethod

from six import add_metaclass


log = logging.getLogger(__name__)


@add_metaclass(ABCMeta)
class BasePlatform(object):
    """
    Base platform engine class.

    This class represent the base interface that topology engines require to
    implement.

    See the :doc:`Plugins Development Guide </plugins>` for reference.

    :param str timestamp: Timestamp in ISO 8601 format.
    :param nmlmanager: Manager holding the NML namespace.
    :type nmlmanager: :class:`pynml.manager.NMLManager`
    """

    @abstractmethod
    def __init__(self, timestamp, nmlmanager):
        pass

    @abstractmethod
    def pre_build(self):
        """
        Preamble hook.

        This hook is called at the first stage of the topology build.
        Use it to setup any requirements your platform has.
        """

    @abstractmethod
    def add_node(self, node):
        """
        Add a node to your platform.

        This method receives a NML node and must return a new Engine Node
        subclass of :class:`BaseNode` that implements the communication
        mechanism with that node.

        :param node: The specification NML node to add to the platform.
        :type node: :class:`pynml.nml.Node`
        :rtype: BaseNode
        :return: Platform specific communication node.
        """

    @abstractmethod
    def add_biport(self, node, biport):
        """
        Add a bidirectional port to the given node.

        This method receives a NML node and a NML bidirectional port that
        specifies the port required to be added. This methods returns nothing
        as the topology users don't expect to deal directly with ports. All
        interaction is done with the nodes.

        :param node: The specification NML node owner of the port.
        :type node: :class:`pynml.nml.Node`
        :param biport: The specification NMP bidirectional port to add to the
         platform.
        :type biport: :class:`pynml.nml.BidirectionalPort`
        :rtype: str
        :return: Real name of the port in the node. This real will be used to
         populate an internal map between the specification name of the port
         and the real name, which can be used to reference the port in commands
         without knowing it's real name in advance.
        """

    @abstractmethod
    def add_bilink(self, nodeport_a, nodeport_b, bilink):
        """
        Add a bidirectional link between given nodes and ports.

        This method receives two tuples with the NML nodes and bidirectional
        ports associated with the bidirectional link and also the NML
        bidirectional link object that specifies the link required to be added
        to the platform. This methods returns nothing as the topology users
        don't expect to deal directly with links. All interaction is done with
        the nodes.

        :param nodeport_a: A tuple (Node, BiPort) of the first endpoint.
        :type nodeport_a:
         (:class:`pynml.nml.Node`, :class:`pynml.nml.BidirectionalPort`)
        :param nodeport_b: A tuple (Node, BiPort) of the second endpoint.
        :type nodeport_b:
         (:class:`pynml.nml.Node`, :class:`pynml.nml.BidirectionalPort`)
        :param bilink: The specification NMP bidirectional link to add to the
         platform.
        :type biport: :class:`pynml.nml.BidirectionalLink`
        """

    @abstractmethod
    def post_build(self):
        """
        Postamble hook.

        This hook is called at the last stage of the topology build.
        Use it to setup any final requirements or start any service before
        using the topology.
        """

    @abstractmethod
    def destroy(self):
        """
        Platform destruction hook.

        Use this to remove all elements from the platform, perform any clean
        and return resources.
        """

    @abstractmethod
    def rollback(self, stage, enodes, exception):
        """
        Platform rollback hook.

        This is called when the build fails, possibly by an exception raised in
        previous hooks.

        :param str stage: The stage the build failed. One of:
         - ``pre_build``.
         - ``add_node``.
         - ``add_biport``.
         - ``add_bilink``.
         - ``post_build``.
        :param enodes: Engine nodes already registered.
        :type enodes: OrderedDict of subclasses of :class:`BaseNode`
        :param Exception exception: The exception caught during build failure.
        """

    @abstractmethod
    def relink(self, link_id):
        """
        Relink back a link specified in the topology.

        :param str link_id: Link identifier to be recreated.
        """

    @abstractmethod
    def unlink(self, link_id):
        """
        Unlink (break) a link specified in the topology.

        :param str link_id: Link identifier to be undone.
        """


__all__ = ['BasePlatform']



#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Common utilities for communication libraries.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division


def stateprovider(stateclass, statename=None, initfunc=None):
    """
    Decorator to inject a library state into an enode.

    :param class stateclass: Class that will hold all the state / attribute of
     the library functions.
    :param str statename: Attribute name to save the instance of the state
     class in the engine node (enode). By default, the name of the state class
     is used like ``_lib_state_<stateclass>`` in lower case.
    :param function initfunc: A function to create the instance the state
     class. It receives as arguments the enode and the state class and MUST
     return and instance of that class already initialized. If ``None``, the
     default, this decorator will create an instance of the state class without
     any arguments or keyword arguments.

    Basic usage:

    ::

       from topology.libraries.utils import stateprovider

       class MyState(object):
           def __init__(self):
               self.my_variable_one = 100

       stateprovider(MyState)
       def my_library_function(enode, state, arg1, arg2, keyword_arg1=None):
           print(state.my_variable_one)
           state.my_variable_one += 100

        __all__ = ['my_library_function']

    In the above example the variable ``my_variable_one`` will be bound to the
    ``enode`` calling the library functions, and each call of that library
    function will increase it's value by 100.
    """

    # Set a default statename
    if statename is None:
        statename = '_lib_state_'.format(stateclass.__name__.lower())

    def decorator(func):
        def replacement(enode, *args, **kwargs):

            state = getattr(enode, statename, None)

            # Create an instance of the state class
            if state is None:

                if initfunc is not None:
                    state = initfunc(enode, stateclass)
                else:
                    state = stateclass()

                setattr(enode, statename, state)

            return func(enode, state, *args, **kwargs)

        replacement.__name__ = func.__name__
        replacement.__doc__ = func.__doc__

        return replacement
    return decorator


__all__ = ['stateprovider']



#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Common utilities for engine platforms.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from copy import copy
from inspect import isclass
from traceback import format_exc
from collections import OrderedDict

from pkg_resources import iter_entry_points

from .node import BaseNode


log = logging.getLogger(__name__)


class NodeLoader(object):
    """
    Node loader utility class for platform engines.

    This class allows to load nodes for a platform engine using Python entry
    points.

    :param str engine_name: Name of the engine.
    :param str api_version: Version of the API for that engine node.
    :param class base_class: Base class to check against. Any class specified
     here must comply with the :class:`BaseNode`.
    """

    def __init__(self, engine_name, api_version='1.0', base_class=None):

        self.entrypoint = 'topology_{engine_name}_node_{api_version}'.format(
            engine_name=engine_name, api_version=api_version.replace('.', '')
        )

        self.base_class = base_class or BaseNode
        assert issubclass(self.base_class, BaseNode)

        self._nodes_cache = OrderedDict()

    def __call__(self, cache=True):
        return self.load_nodes(cache=cache)

    def load_nodes(self, cache=True):
        """
        List all available nodes types.

        This function lists all available node types by discovering installed
        plugins registered in the entry point. This can be costly or error
        prone if a plugin misbehave. Because of this a cache is stored after
        the first call.

        :param bool cache: If ``True`` return the cached result. If ``False``
         force reload of all plugins registered for the entry point.
        :rtype: OrderedDict
        :return: An ordered dictionary associating the name of the node type
         and the class (subclass of :class:`topology.platforms.node.BaseNode`)
         implementing it.
        """

        # Return cached value if call is repeated
        if cache and self._nodes_cache:
            return copy(self._nodes_cache)

        # Add built-in node types
        available = OrderedDict()

        # Iterate over entry points
        for ep in iter_entry_points(group=self.entrypoint):

            name = ep.name

            try:
                node = ep.load()
            except:
                log.error(
                    'Unable to load node from plugin {}'.format(name)
                )
                log.debug(format_exc())
                continue

            if not isclass(node) or not issubclass(node, self.base_class):
                log.error(
                    'Ignoring node "{}" as it doesn\'t '
                    'match the required interface: '
                    'Node not a subclass of {}.'.format(
                        name, self.base_class.__name__
                    )
                )
                continue

            available[name] = node

        self._nodes_cache = available
        return copy(self._nodes_cache)


__all__ = ['NodeLoader']



#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Deprecated module that once included the classes:

- :class:`topology.platforms.platform.BasePlatform`: Base abstract class to
  implement an engine platform.
- :class:`topology.platforms.node.BaseNode` and
  :class:`topology.platforms.node.CommonNode`: Base abstract classes to
  implement nodes for such engine.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

from warnings import warn

from .platform import BasePlatform
from .node import BaseNode, CommonNode


class DeprecatedTopologyModule(UserWarning):
    pass


warn(
    'This module is deprecated, please change all imports from it.\n'
    'topology.platforms.base.BasePlatform => '
    'topology.platforms.platform.BasePlatform\n'
    'topology.platforms.base.BaseNode => '
    'topology.platforms.node.BaseNode\n'
    'topology.platforms.base.CommonNode => '
    'topology.platforms.node.CommonNode\n',
    category=DeprecatedTopologyModule,
)

__all__ = ['BasePlatform', 'BaseNode', 'CommonNode']
__api__ = []



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
topology module for interactive mode.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from atexit import register
from inspect import ismodule
from traceback import format_exc
from os.path import join, expanduser, isfile


log = logging.getLogger(__name__)


def catch_exc(func):
    """
    Decorator for in-shell exception handling.

    This is particularly required for ``readline`` hooks, as they run in the
    C library and cannot catch any Python exception ocurring in them.
    """
    def catcher(*args):
        try:
            return func(*args)
        except:
            log.error(format_exc())
    return catcher


class NamespaceCompleter(object):
    """
    Readline completer using a dictionary namespace.
    """

    def __init__(self, ns):
        self.matches = []
        self.ns = ns

    def search_tree(self, root, path):
        """
        Search for a node in a Python object as if it were a tree: namespaces,
        modules, objects, method, attributes, etc.

        :param root: The root Python object.
        :param list path: List of names representing the path.
        :rtype: list
        :return: The subnode or None if not found.
        """
        node = root

        for p in path:

            attrs = self.dict_attributes(node)

            if p not in attrs:
                node = None
                break

            node = attrs[p]

        return node

    def dict_attributes(self, element):
        """
        Type aware attributes to dictionary function.

        :param element: Python object to list attributes.
        :rtype: dict
        :return: A dict of the element attributes.
        """
        if element is not None:

            # Element is a dictionary, return itself
            if isinstance(element, dict):
                return element

            # Element is a module as has a __all__ defined
            if ismodule(element) and hasattr(element, '__all__'):
                return {
                    k: v for k, v in vars(element).items()
                    if k in element.__all__
                }

            # The element can be converted to dict
            if hasattr(element, '__dict__'):
                return vars(element)

        return {}

    def format_matches(self, path, attrs, prefixed):
        """
        Format a list of attributes to be shown in autocompleter.

        :param path: List of string representing the path to the node.
        :param attrs: List of attributes of the node.
        :param prefixed: Prefix attributes must have in order to be shown.
        :rtype: list
        :return: Formatted list of attributes.
        """
        # Filter super private attributes
        attrs = [a for a in attrs if not a.startswith('__')]

        # Filter attributes if a prefix is required
        if prefixed:
            attrs = [a for a in attrs if a.startswith(prefixed)]

        return ['.'.join(path + [a]) for a in attrs]

    @catch_exc
    def complete(self, text, state):

        if state == 0:  # On first trigger, build possible matches

            path = text.split('.')
            search = path.pop()

            subnode = self.search_tree(self.ns, path)
            attributes = sorted(self.dict_attributes(subnode))
            self.matches = self.format_matches(path, attributes, search)

        # Return match indexed by state
        try:
            return self.matches[state]
        except IndexError:
            return None


def interact(mgr):
    """
    Shell setup function.

    This function will setup the library, create the default namespace with
    shell symbols, setup the history file, the autocompleter and launch a shell
    session.
    """
    print('Engine nodes available for communication:')
    print('    {}'.format(', '.join(mgr.nodes.keys())))

    # Create and populate shell namespace
    ns = {
        'topology': mgr
    }
    for key, enode in mgr.nodes.items():
        ns[key] = enode

    # Configure readline, history and autocompleter
    import readline

    histfile = join(expanduser('~'), '.topology_history')
    if isfile(histfile):
        try:
            readline.read_history_file(histfile)
        except IOError:
            log.error(format_exc())
    register(readline.write_history_file, histfile)

    completer = NamespaceCompleter(ns)
    readline.set_completer(completer.complete)
    readline.parse_and_bind('tab: complete')

    # Python's Interactive
    from code import interact as pyinteract
    pyinteract('', None, ns)


__all__ = ['NamespaceCompleter', 'interact']



#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Common helpers communication library implementation.

This library defines some helpers as a communication library.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division


def assert_batch(enode, commands, replace=None, shell=None):
    """
    Execute a batch of return-less template commands in the enode shell.

    :param enode:
    :type: topology.platforms.node.BaseNode
    :param str commands: Multiline docstring template with the return-less
     commnads to execute and assert.
    :param dict replace: Namespace to replace tokens in commands.
    :param str shell: Shell name to execute the batch.
    """
    assert commands

    if replace is not None:
        assert isinstance(replace, dict)
        commands = commands.format(**replace)

    for cmd in commands.splitlines():

        # Ignore empty lines in batch
        cmd = cmd.strip()
        if not cmd:
            continue

        # Assert that commands return nothing
        response = enode(cmd, shell=shell)
        assert not response


__all__ = [
    'assert_batch'
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
topology service objects module.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

from logging import getLogger


log = getLogger()


class BaseService(object):
    """
    Base class used to describe a service.

    It is mainly used to store context variables.

    :param str name: Name of the service.
    :param int port: Port of the service.
    :param str protocol: Protocol used by the service.

    :var str name: Given name in the constructor.
    :var int port: Given port in the constructor.
    :var str protocol: Given protocol in the constructor.
    :var str address: IP or FQDN of the node exposing this service. This is
     set by the engine node.
    """
    def __init__(self, name, port, protocol='tcp'):
        self.name = name
        self.port = port
        self.protocol = protocol
        self.address = None

    def __str__(self):
        return '{protocol}://{address}:{port}/'.format(**self.__dict__)


__all__ = ['BaseService']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Application entry point module.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import sys
import logging
from atexit import register
from os.path import join, basename, splitext

from six import StringIO

from . import __version__
from .manager import TopologyManager
from .interact import interact
from .parser import find_topology_in_python
from .injection import parse_attribute_injection
from .logging import manager as logmanager


log = logging.getLogger(__name__)


def main(args):
    """
    Application main function.

    :param args: An arguments namespace.
    :type args: :py:class:`argparse.Namespace`
    :return: Exit code.
    :rtype: int
    """
    print('Starting Network Topology Framework v{}'.format(__version__))

    # Setup framework logging
    logmanager.logging_context = None
    if args.log_dir:
        logmanager.logging_directory = args.log_dir

    # Parse attributes injection file
    injected_attr = None
    if args.inject is not None:
        injection_spec = parse_attribute_injection(args.inject)
        injected_attr = injection_spec.get(args.topology, None)

    # Create manager
    mgr = TopologyManager(args.platform)

    # Read topology
    if args.topology.endswith('.py'):

        topology = find_topology_in_python(args.topology)

        if topology is None:
            log.error(
                'TOPOLOGY variable could not be found in file {}'.format(
                    args.topology
                )
            )
            return 1

        mgr.parse(topology, inject=injected_attr)
    else:
        with open(args.topology, 'r') as fd:
            mgr.parse(fd.read(), inject=injected_attr)

    print('Building topology, please wait...')

    # Capture stdout to hide commands during build
    if not args.show_build_commands:
        sys.stdout = StringIO()

    # Build topology
    mgr.build()

    # Restore stdout after build
    if not args.show_build_commands:
        sys.stdout = sys.__stdout__

    # Start interactive mode if required
    if not args.non_interactive:

        # Register unbuild
        def unbuild():
            print('Unbuilding topology, please wait...')
            mgr.unbuild()
        register(unbuild)

        interact(mgr)

    # Plot and export topology
    module = splitext(basename(args.topology))[0]
    if args.plot_dir:
        plot_file = join(
            args.plot_dir, '{}.{}'.format(module, args.plot_format)
        )
        mgr.nml.save_graphviz(
            plot_file, keep_gv=True
        )

    # Export topology as NML
    if args.nml_dir:
        nml_file = join(
            args.nml_dir, '{}.xml'.format(module)
        )
        mgr.nml.save_nml(
            nml_file, pretty=True
        )

    return 0


__all__ = ['main']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
topology shell api module.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

from warnings import warn
from re import sub as regex_sub
from collections import OrderedDict
from abc import ABCMeta, abstractmethod

from six import add_metaclass
from pexpect import spawn as Spawn  # noqa

from topology.logging import get_logger


TERM_CODES_REGEX = r'\x1b[E|\[](\?)?([0-9]{1,2}(;[0-9]{1,2})?)?[m|K|h|H|r]?'
"""
Regular expression to match terminal control codes.

A terminal control code is a special sequence of characters that sent by some
applications to control certain features of the terminal. It is responsibility
of the terminal application (driver) to interpret those control codes.
However, when using pexpect, the incoming buffer will still have those control
codes and will not interpret or do anything with them other than store them.
This regular expression allows to remove them as they are unneeded for the
purpose of executing commands and parsing their outputs
(unless proven otherwise).

``\\x1b``
  Match prefix that indicates the next characters are part of a terminal code
  string.
``[E|\\[]``
  Match either ``E`` or ``[``.
``(\\?)?``
  Match zero or one ``?``.
``([0-9]{1,2}``
  Match 1 or 2 numerical digits.
``(;[0-9]{1,2})?``
  Match zero or one occurences of the following pattern: ``;`` followed by 1
  or 2 numerical digits.
``)?``
  Means the pattern composed by the the last 2 parts above can be found zero or
  one times.
``[m|K|h|H|r]?``
  Match zero or one occurrences of either ``m``, ``K``, ``h``, ``H`` or ``r``.
"""


class DeprecatedTopologyFunction(UserWarning):
    pass


class NonExistingConnectionError(Exception):
    """
    Exception raised by the shell API when trying to use a non-existent
    connection.
    """
    def __init__(self, connection):
        super(Exception, self).__init__(
            self, 'Non-existing connection {}.'.format(connection)
        )


class AlreadyConnectedError(Exception):
    """
    Exception raised by the shell API when trying to create a connection
    already created.
    """
    def __init__(self, connection):
        super(Exception, self).__init__(
            self, '{} is already connected.'.format(connection)
        )


class AlreadyDisconnectedError(Exception):
    """
    Exception raised by the shell API when trying to disconnect an already
    disconnected connection.
    """
    def __init__(self, connection):
        super(Exception, self).__init__(
            self, '{} is already disconnected.'.format(connection)
        )


class DisconnectedError(Exception):
    """
    Exception raised by the shell API when trying to perform an operation on a
    disconnected connection.
    """
    def __init__(self, connection):
        super(Exception, self).__init__(
            self, '{} is disconnected.'.format(connection)
        )


@add_metaclass(ABCMeta)
class BaseShell(object):
    """
    Base shell class for Topology nodes.

    This class represents a base interface for a Topology node shell. This
    shell is expected to be an interactive shell, where an expect-like
    mechanism is to be used to find a terminal prompt that signals the end of
    the terminal response to a command sent to it.

    Shells of this kind also represent an actual shell in the node. This means
    that one of these objects is expected to exist for every one of the shells
    in the node. These shells are accessible through a *connection* that may be
    implemented using a certain command (``telnet``, or ``ssh``, for example).
    It may be possible to have several connections to the same shell, so these
    shells support multiple connections.

    The behavior that these connections must follow is this one:

    #. New connections to the shell are created by calling the ``connect``
       command of the shell with the name of the connection to be created
       defined in its ``connection`` attribute.
    #. Connections can be disconnected by calling the ``disconnect`` command of
       the shell.
    #. Connections can be either connected or disconnected.
    #. Existing disconnected connections can be connected again by calling the
       ``connnect`` command of the shell.
    #. An attempt to connect an already connected shell or to disconnect an
       already disconnected shell will raise an exception.
    #. These shells will have a *default* connection that will be used if the
       ``connection`` attribute of the shell methods is set to ``None``.
    #. The default connection will be set to ``None`` when the shell is
       initialized.
    #. If the default connection is ``None`` and the ``connect`` command is
       called, it will set the default connection as the connection used in
       this call. Also, if the previous condition is true and if this
       connection attribute is set to ``None``, the name of the default
       connection will be set to '0'.
    #. Every time any operation with the shell is attempted, a connection needs
       to be specified, if this connection is specified to be ``None``, the
       *default* connection of the shell will be used.
    #. If any method other than ``connect`` and ``send_command`` is called with
       a connection that is not defined yet, an exception will be raised.
    #. If ``auto_connect`` is True, calling ``send_command`` with a
       disconnected or nonexistent connection will create and use a new
       connection. An exception will be raised otherwise.

    The behavior of these operations is defined in the following methods,
    implementations of this class are expected to behave as defined here.

    Be aware that the method ``_register_shell`` of the node will add two
    attributes to the shell objects:

    #. ``_name``: name of the shell, the matching key of the node's dictionary
       of shells
    #. ``_node``: node object that holds the shell object

    The ``_register_shell`` method is usually called in the node's'
    ``__init__``.
    """

    @property
    def default_connection(self):
        raise NotImplementedError('default_connection')

    @default_connection.setter
    def default_shell(self, value):
        raise NotImplementedError('default_connection.setter')

    @abstractmethod
    def send_command(
        self, command,
        matches=None, newline=True,
        timeout=None, connection=None, silent=False
    ):
        """
        Send a command to the shell.

        :param str command: Command to be sent to the shell.
        :param list matches: List of strings that may be matched by the shell
         expect-like mechanism as prompts in the command response.
        :param bool newline: True to append a newline at the end of the
         command, False otherwise.
        :param int timeout: Amount of time to wait until a prompt match is
         found in the command response.
        :param str connection: Name of the connection to be used to send the
         command. If not defined, the default connection will be used.
        :param bool silent: True to call the connection logger, False
         otherwise.
        """

    @abstractmethod
    def get_response(self, connection=None, silent=False):
        """
        Get a response from the shell connection.

        This method can be used to add extra processing to the shell response
        if needed, cleaning up terminal control codes is an example.

        :param str connection: Name of the connection to be used to get the
         response from. If not defined, the default connection will be used.
        :param bool silent: True to call the connection logger, False
         otherwise.
        :rtype: str
        :return: Shell response to the previously sent command.
        """

    @abstractmethod
    def is_connected(self, connection=None):
        """
        Shows if the connection to the shell is active.

        :param str connection: Name of the connection to check if connected. If
         not defined, the default connection will be checked.
        :rtype: bool
        :return: True if there is an active connection to the shell, False
         otherwise.
        """

    @abstractmethod
    def connect(self, connection=None):
        """
        Creates a connection to the shell.

        :param str connection: Name of the connection to be created. If not
         defined, an attempt to create the default connection will be done. If
         the default connection is already connected, an exception will be
         raised. If defined in the call but no default connection has been
         defined yet, this connection will become the default one. If the
         connection is already connected, an exception will be raised.
        """

    @abstractmethod
    def disconnect(self, connection=None):
        """
        Terminates a connection to the shell.

        :param str connection: Name of the connection to be disconnected. If
         not defined, the default connection will be disconnected. If the
         default connection is disconnected, no attempt will be done to define
         a new default connection, the user will have to either create a new
         default connection by calling ``connect`` or by defining another
         existing connection as the default one.
        """

    def execute(self, command, connection=None):
        """
        Executes a command.

        If the default connection is not defined, or is disconnected, an
        exception will be raised.

        This is just a convenient method that sends a command to the shell
        using send_command and returns its response using get_response.

        :param str command: Command to be sent.
        :param str connection: Connection to be used to execute this command.
        :rtype: str
        :return: Shell response to the command being sent.
        """
        self.send_command(command, connection=connection)
        return self.get_response(connection=connection)

    def __call__(self, command, connection=None):
        return self.execute(command, connection=connection)

    def _setup_shell(self, connection=None):
        """
        Method called by subclasses that will be triggered after matching the
        initial prompt.

        This method must do a call to expect before it does one to any sending
        call (like send or sendline) and it must do a sending call after the
        last call to expect.

        :param str connection: Name of the connection to be set up. If not
         defined, the default connection will be set up.
        """

    def _register_node(self, node_identifier, shell_name):
        """
        Register the node identifier and the assigned shell name on that node
        in the shell for logging back reference.

        :param str node_identifier: The identifier of the owner node.
        :param str shell_name: Shell name given in the node to this shell.
        """

    def _register_loggers(
        self, node, shell, command_logger=None, response_logger=None
    ):
        """
        Register logger functions for command executions and responses.

        .. warning::

           This function will be removed in future releases.

        :param :class:`topology.base.BaseNode` node: Node that holds this
         shell.
        :param str shell: Name of the shell to show in the sent command log.
        :param function command_logger: Function that logs the command being
         sent. If set to None, the node's _log_command will be used.
        :param function response_logger: Function that logs the command
         response. If set to None, the node's _log_response will be used.
        """


@add_metaclass(ABCMeta)
class PExpectShell(BaseShell):
    """
    Implementation of the BaseShell class using pexpect.

    This class provides a convenient implementation of the BaseShell using
    the pexpect package. The only thing needed for child classes is to define
    the command that will be used to connect to the shell.

    See :class:`BaseShell`.

    :param str prompt: Regular expression that matches the shell prompt.
    :param str initial_command: Command that is to be sent at the beginning of
     the connection.
    :param str initial_prompt: Regular expression that matches the initial
     prompt. This value will be used to match the prompt before calling private
     method ``_setup_shell()``.
    :param str password: Password to be sent at the beginning of the
     connection.
    :param str password_match: Regular expression that matches a password
     prompt.
    :param str prefix: The prefix to be prepended to all commands sent to this
     shell.
    :param int timeout: Default timeout to use in send_command.
    :param str encoding: Character encoding to use when decoding the shell
     response.
    :param bool try_filter_echo: On platforms that doesn't support some way of
     turning off the echo of the command try to filter the echo from the output
     by removing the first line of the output if it match the command.
    :param auto_connect bool: Enable the automatic creation of a connection
     when ``send_command`` is called if the connection does not exist or is
     disconnected.
    :param dict spawn_args: Arguments to be passed to the Pexpect spawn
     constructor. If this is left as ``None``, then
     ``env={'TERM': 'dumb'}, echo=False`` will be passed as keyword
     arguments to the spawn constructor.
    :param str errors: Handling of decoding errors in ``get_response``. The
     values available are the same ones that the ``decode`` method of a bytes
     object expects in its ``error`` keyword argument. Defaults to ``ignore``.
    """

    def __init__(
            self, prompt,
            initial_command=None, initial_prompt=None,
            user=None, user_match='[uU]ser:',
            password=None, password_match='[pP]assword:',
            prefix=None, timeout=None, encoding='utf-8',
            try_filter_echo=True, auto_connect=True,
            spawn_args=None, errors='ignore', **kwargs):

        self._connections = OrderedDict()
        self._default_connection = None

        self._initial_command = initial_command
        self._prompt = prompt
        self._initial_prompt = initial_prompt
        self._user = user
        self._user_match = user_match
        self._password = password
        self._password_match = password_match
        self._prefix = prefix
        self._timeout = timeout or -1
        self._encoding = encoding
        self._try_filter_echo = try_filter_echo
        self._auto_connect = auto_connect
        self._command_logger = None
        self._response_logger = None
        self._node_identifier = None
        self._shell_name = None
        self._errors = errors

        # Doing this to avoid having a mutable object as default value in the
        # arguments.
        if spawn_args is None:
            self._spawn_args = {'env': {'TERM': 'dumb'}, 'echo': False}
        else:
            self._spawn_args = spawn_args

        self._last_command = None

        # Set the initial prompt not specified
        if self._initial_prompt is None:
            self._initial_prompt = prompt

        super(PExpectShell, self).__init__(**kwargs)

    @property
    def default_connection(self):
        return self._default_connection

    @default_connection.setter
    def default_connection(self, connection):
        if connection not in self._connections:
            raise NonExistingConnectionError(connection)
        self._default_connection = connection

    @abstractmethod
    def _get_connect_command(self):
        """
        Get the command to be used when connecting to the shell.

        This must be defined by any child class as the return value of this
        function will define all the connection details to use when creating a
        connection to the shell. It will be used usually in conjunction with
        other shell attributes to define the exact values to use when creating
        the connection.

        :rtype: str
        :return: The command to be used when connecting to the shell.
        """

    def _get_connection(self, connection):
        """
        Get the pexpect object for the given connection name.

        :param str connection: Name of the connection.
        :rtype: :class:`pexpect.spawn`
        :return: The PExpect spawn process that handles the connection.
        """
        connection = connection or self._default_connection or '0'

        if connection not in self._connections:
            raise NonExistingConnectionError(connection)

        return self._connections[connection]

    def send_command(
        self, command,
        matches=None, newline=True,
        timeout=None, connection=None, silent=False
    ):
        """
        See :meth:`BaseShell.send_command` for more information.
        """
        # If auto connect is false, fail if:
        # 1. Connection is missing
        # 2. Connection is disconnected
        if not self._auto_connect:
            if not self.is_connected(connection):
                raise DisconnectedError(connection)

        # If auto connect is true, always reconnect unless:
        # 1. Connection already exists, and
        # 2. Connection is connected
        else:
            try:
                if not self.is_connected(connection):
                    self.connect(connection)
            except NonExistingConnectionError:
                self.connect(connection)

        spawn = self._get_connection(connection)

        # Create possible expect matches
        if matches is None:
            matches = [self._prompt]

        # Append prefix if required
        if self._prefix is not None:
            command = '{}{}'.format(self._prefix, command)

        # Save last command in cache to allow to remove echos in get_response()
        self._last_command = command

        # Send line and expect matches
        if newline:
            spawn.sendline(command)
        else:
            spawn.send(command)

        # Log log_send_command
        if not silent:
            spawn._connection_logger.log_send_command(
                command, matches, newline, timeout
            )

        # Expect matches
        if timeout is None:
            timeout = self._timeout

        match_index = spawn.expect(
            matches, timeout=timeout
        )
        return match_index

    def get_response(self, connection=None, silent=False):
        """
        See :meth:`BaseShell.get_response` for more information.
        """
        # Get connection
        spawn = self._get_connection(connection)

        # Convert binary representation to unicode using encoding
        text = spawn.before.decode(
            encoding=self._encoding, errors=self._errors
        )

        # Remove leading and trailing whitespaces and normalize newlines
        text = text.strip().replace('\r', '')

        # Remove control codes
        text = regex_sub(TERM_CODES_REGEX, '', text)

        # Split text into lines
        lines = text.splitlines()

        # Delete buffer with output right now, as it can be very large
        del text

        # Remove echo command if it exists
        if self._try_filter_echo and \
                lines and self._last_command is not None \
                and lines[0].strip() == self._last_command.strip():
            lines.pop(0)

        response = '\n'.join(lines)

        # Log response
        if not silent:
            spawn._connection_logger.log_get_response(response)

        return response

    def is_connected(self, connection=None):
        """
        See :meth:`BaseShell.is_connected` for more information.
        """
        # Get connection
        spawn = self._get_connection(connection)
        return spawn.isalive()

    def connect(self, connection=None):
        """
        See :meth:`BaseShell.connect` for more information.
        """
        connection = connection or self._default_connection or '0'

        if connection in self._connections and self.is_connected(connection):
            raise AlreadyConnectedError(connection)

        # Inject framework logger to the spawn object
        spawn_args = {
            'logfile': get_logger(
                OrderedDict([
                    ('node_identifier', self._node_identifier),
                    ('shell_name', self._shell_name),
                    ('connection', connection)
                ]),
                category='pexpect'
            ),
        }

        # Create a child process
        spawn_args.update(self._spawn_args)

        spawn = Spawn(
            self._get_connect_command().strip(),
            **spawn_args
        )

        # Add a connection logger
        # Note: self._node and self._name were added to this shell in the
        #       node's call to its _register_shell method.
        spawn._connection_logger = get_logger(
            OrderedDict([
                ('node_identifier', self._node_identifier),
                ('shell_name', self._shell_name),
                ('connection', connection)
            ]),
            category='connection'
        )

        self._connections[connection] = spawn

        try:
            def expect_sendline(prompt, command):
                if command is not None:
                    spawn.expect(
                        prompt, timeout=self._timeout
                    )
                    spawn.sendline(command)

            # If connection is via user
            expect_sendline(self._user_match, self._user)

            # If connection is via password
            expect_sendline(self._password_match, self._password)

            # If connection is via initial command
            expect_sendline(self._initial_prompt, self._initial_command)

            # Setup shell before using it
            self._setup_shell(connection)

            # Wait for command response to match the prompt
            spawn.expect(
                self._prompt, timeout=self._timeout
            )

        except:
            # Always remove a bad connection if it failed
            del self._connections[connection]
            raise

        # Set connection as default connection if required
        if self.default_connection is None:
            self.default_connection = connection

    def disconnect(self, connection=None):
        """
        See :meth:`BaseShell.disconnect` for more information.
        """
        # Get connection
        spawn = self._get_connection(connection)
        if not spawn.isalive():
            raise AlreadyDisconnectedError(connection)
        spawn.close()

    def _register_node(self, node_identifier, shell_name):
        self._node_identifier = node_identifier
        self._shell_name = shell_name

    def _register_loggers(
        self, node, shell, command_logger=None, response_logger=None
    ):

        warn(
            'This method is deprecated, instead of calling _register_loggers '
            'in the shell object, please just call _register_shell in the'
            ' node that holds the shell object.',
            category=DeprecatedTopologyFunction
        )

        self._shell = shell

        if command_logger is None:
            self._command_logger = node._log_command
        else:
            self._command_logger = command_logger

        if response_logger is None:
            self._response_logger = node._log_response
        else:
            self._response_logger = response_logger


class PExpectBashShell(PExpectShell):
    """
    Custom shell class for Bash.

    This custom base class will setup the prompt ``PS1`` to the
    ``FORCED_PROMPT`` value of the class and will disable the echo of the
    device by issuing the ``stty -echo`` command. All this is done in the
    ``_setup_shell()`` call, which is overriden by this class.
    """
    FORCED_PROMPT = '@~~==::BASH_PROMPT::==~~@'

    def __init__(
            self,
            initial_prompt='\w+@.+:.+[#$] ', try_filter_echo=False,
            **kwargs):

        super(PExpectBashShell, self).__init__(
            PExpectBashShell.FORCED_PROMPT,
            initial_prompt=initial_prompt,
            try_filter_echo=try_filter_echo,
            **kwargs
        )

    def _setup_shell(self, connection=None):
        """
        Overriden setup function that will disable the echo on the device on
        the shell and set a pexpect-safe prompt.
        """
        spawn = self._get_connection(connection)

        # Wait initial prompt
        spawn.expect(
            self._initial_prompt, timeout=self._timeout
        )

        # Remove echo
        spawn.sendline('stty -echo')
        spawn.expect(
            self._initial_prompt, timeout=self._timeout
        )

        # Change prompt to a pexpect secure prompt
        spawn.sendline(
            'export PS1={}'.format(PExpectBashShell.FORCED_PROMPT)
        )
        self._prompt = PExpectBashShell.FORCED_PROMPT


class ShellContext(object):
    """
    Context Manager class for default shell swapping.

    This object will handle the swapping of the default shell when in and out
    of the context.

    :param BaseNode node: Node to default shell to swap.
    :param str shell_to_use: Shell to use during the context session.
    """

    def __init__(self, node, shell_to_use):
        self._node = node
        self._shell_to_use = shell_to_use
        self._default_shell = node.default_shell

    def __enter__(self):
        self._node.default_shell = self._shell_to_use
        return self._node.get_shell(self._default_shell)

    def __exit__(self, type, value, traceback):
        self._node.default_shell = self._default_shell


__all__ = [
    'TERM_CODES_REGEX',
    'NonExistingConnectionError',
    'AlreadyConnectedError',
    'AlreadyDisconnectedError',
    'BaseShell',
    'PExpectShell',
    'PExpectBashShell',
    'ShellContext'
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Parser attributes injection files.

This module defines a set of functions that implement logic to parse a
attributes injection file and extract from it the injection specification that
allows to modify attributes of entities in different topologies.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

from re import match
from glob import glob
from json import loads
from os import getcwd, walk
from fnmatch import fnmatch
from logging import getLogger
from traceback import format_exc
from collections import OrderedDict
from os.path import isabs, join, abspath, isfile, basename

from six import iteritems

from .parser import parse_txtmeta, find_topology_in_python


log = getLogger(__name__)


def parse_attribute_injection(injection_file, search_paths=None):
    """
    Parses a attributes injection file into an attribute injection dictionary.

    An attribute injection file is a JSON file that specifies a list of
    injection specifications. Each specification allows to list the files to
    modify and the modifiers to apply. Each modifier specify what entities
    (devices, ports or links) will be modified and which attributes will be
    injected. The file support willcards and attributes matching.

    The file format is as follows:

    ::

        [
            {
                "files": ["pathto/*", "another/foo*.py"],
                "modifiers": [
                    {
                        "nodes": ["sw1", "type=host", "sw3"],
                        "attributes": {
                            "image": "image_for_sw1_sw3_hs1_hs2",
                            "hardware": "hardware_for_sw1_sw3_hs1_hs2"
                        }
                    },
                    {
                        "nodes": ["sw4"],
                        "attributes": {
                            "image": "image_for_sw4"
                        }
                    },
                    ... # More modifiers
                ]
            },
            ... # More injection specifications
        ]

    :param str injection_file: Path for the attribute injection file.
    :param list search_paths: Paths to search for files when the file match is
     relative in the injection file.
     If ``None`` (the default), the current working directory is used.
    :return: An ordered dictionary with the attributes to inject of the form:

     ::

        {
            '/abs/path/to/suite.py': {
                'sw1': {
                    'image': 'image_for_sw1_sw3_hs1_hs2',
                    'hardware': 'hardware_for_sw1_sw3_hs1_hs2'
                },
                'sw3': {
                    'image': 'image_for_sw1_sw3_hs1_hs2',
                    'hardware': 'hardware_for_sw1_sw3_hs1_hs2'
                }
            }
        }

    :rtype: `collections.OrderedDict`
    """
    # Define search paths
    if search_paths is None:
        search_paths = [abspath(getcwd())]
    log.debug('Injection search paths: {}'.format(search_paths))

    # Expand search paths recursively to include all subfolders
    def subfolders(search_path):
        result = []
        for root, dirs, files in walk(search_path, topdown=True,
                                      followlinks=True):
            # Ignore hidden folders
            dirs[:] = [d for d in dirs if not d.startswith('.')]
            result.extend([join(root, directory) for directory in dirs])
        return result

    paths_to_expand = list(search_paths)
    for root in paths_to_expand:
        children = subfolders(root)
        search_paths.extend(children)

    # Make search paths unique
    uniques = []
    for path in search_paths:
        if path not in uniques:
            uniques.append(path)
    search_paths = uniques
    log.debug('Expanded injection search paths: {}'.format(search_paths))

    # Read injection file
    with open(injection_file) as fd:
        injection_spec = loads(fd.read())

    result = OrderedDict()

    # Iterate all specifications, expand files and fill return dictionary
    for spec in injection_spec:
        for filename in expand_files(spec['files'], search_paths):

            if filename not in result:
                result[filename] = OrderedDict()

            # Each specification have several "modifiers" associated to it.
            # Those modifiers hold the nodes whose attributes are to be
            # modified.
            for modifier in spec['modifiers']:
                for node in expand_nodes(filename, modifier['nodes']):

                    if node not in result[filename]:
                        result[filename][node] = {}

                    for attribute, value in modifier['attributes'].items():
                        result[filename][node][attribute] = value

    log.debug('Attribute injection interpreted dictionary:')
    log.debug(result)

    return result


def expand_files(files_definitions, search_paths):
    """
    Expands a list of files definitions into the matching files paths.

    A file definition is a string that can match none, one or more files
    (by using wildcards). It can be an absolute path, or a relative path from
    the search paths. For example:

    ::

        '/abs/path/to/my*_thing.py'
        'myfile.szn'
        'relative/test_*.py'

    :param list files_definitions: A list of files definitions.
    :param str search_paths: Paths to search for files when the file definition
     is relative.
    :return: A list of files paths.
    """

    expanded_files = []

    for file_definition in files_definitions:

        # File definitions to look for
        lookups = []

        # Determine if suite must be located in suites search path
        if isabs(file_definition):
            lookups.append(file_definition)
        else:
            for search_path in search_paths:
                lookups.append(join(search_path, file_definition))

        # Find all file matches for the suite definition
        matches = []
        for lookup in lookups:
            for filepath in glob(lookup):
                filename = basename(filepath)

                if filepath in expanded_files or not isfile(filepath):
                    continue

                if fnmatch(filename, 'test_*.py') or \
                        fnmatch(filename, '*.szn'):
                    matches.append(filepath)

        expanded_files.extend(matches)

    return expanded_files


def expand_nodes(filename, nodes_definitions):
    """
    Expands a list of node definitions into the matching node names.

    A node definition is a string that can match none, one or more nodes
    (by using wildcards). It can be an expression for node name matching, or
    for matching all nodes that have an specific attribute value. For example:

    ::

        'nodea'
        'hs*'
        'type=host'

    :param str filename: A filename in which to look for matching nodes.
    :param list nodes_definitions: A list of node definitions.
    :return: A list of matching nodes.
    """

    expanded_nodes = []

    # Grab the topology definition from a file that contains one
    log.debug('Trying to expand nodes in {}'.format(filename))
    if filename.endswith('.py'):
        topology = find_topology_in_python(filename)
        if topology is None:
            log.warning((
                'Skipping node expansion for attribute injection in filename '
                '{} in the lookup path as it does not contain a TOPOLOGY '
                'definition.'
            ).format(filename))
            return []
    else:
        with open(filename, 'r') as fd:
            topology = fd.read().strip()
    log.debug('Found:\n{}'.format(topology))

    # Parse content
    try:
        parsed_topology = parse_txtmeta(topology)
    except:
        log.error((
            'Skipping node expansion for attribute injection in filename '
            '{} in the lookup path as SZN format parsing failed.'
        ).format(filename))
        log.debug(format_exc())
        return []

    for node_definition in nodes_definitions:

        # Check if definition is for attribute matching
        if match(r'(\w+)=(\w+)', node_definition):

            # Build a dummy statement and parse it
            parsed_dummy = parse_txtmeta(
                '[{}] dummy1'.format(node_definition)
            )

            # Extract the attribute name and value
            attribute, value = list(iteritems(
                parsed_dummy['nodes'][0]['attributes']
            ))[0]

            # Look for attribute name matching
            for nodes_group in parsed_topology['nodes']:
                attributes = nodes_group['attributes']
                if attribute in attributes and value == attributes[attribute]:

                    # Retain order, and avoid adding repeated nodes
                    for node in nodes_group['nodes']:
                        if node not in expanded_nodes:
                            expanded_nodes.append(node)
            continue

        # The definition is not attribute matching, but name matching
        for nodes_group in parsed_topology['nodes']:
            for node in nodes_group['nodes']:
                if fnmatch(node, node_definition) and \
                        node not in expanded_nodes:
                    expanded_nodes.append(node)

    return expanded_nodes


__all__ = [
    'parse_attribute_injection'
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Argument management module.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from os import getcwd
from distutils.dir_util import mkpath
from os.path import join, isabs, abspath, isfile

from . import __version__
from .platforms.manager import platforms, DEFAULT_PLATFORM


log = logging.getLogger(__name__)


FORMAT = '%(asctime)s:::%(levelname)s:::%(message)s'
V_LEVELS = {
    0: logging.ERROR,
    1: logging.WARNING,
    2: logging.INFO,
    3: logging.DEBUG,
}


def validate_args(args):
    """
    Validate that arguments are valid.

    :param args: An arguments namespace.
    :type args: :py:class:`argparse.Namespace`
    :return: The validated namespace.
    :rtype: :py:class:`argparse.Namespace`
    """
    level = V_LEVELS.get(args.verbose, logging.DEBUG)
    logging.basicConfig(format=FORMAT, level=level)

    log.debug('Raw arguments:\n{}'.format(args))

    # Verify topology file exists
    if not isfile(args.topology):
        log.error('No such file : {}'.format(args.topology))
        exit(1)
    args.topology = abspath(args.topology)

    # Determine plot directory and create it if required
    if args.plot_dir:
        if not isabs(args.plot_dir):
            args.plot_dir = join(abspath(getcwd()), args.plot_dir)
        mkpath(args.nml_dir)

    # Determine NML export directory and create it if required
    if args.nml_dir:
        if not isabs(args.nml_dir):
            args.nml_dir = join(abspath(getcwd()), args.nml_dir)
        mkpath(args.nml_dir)

    # Verify inject file exists
    if args.inject:
        if not isfile(args.inject):
            log.error('No such file : {}'.format(args.inject))
            exit(1)
        args.inject = abspath(args.inject)

    # Determine log directory and create it if required
    if args.log_dir:
        if not isabs(args.log_dir):
            args.log_dir = join(abspath(getcwd()), args.log_dir)
        mkpath(args.log_dir)

    return args


def parse_args(argv=None):
    """
    Argument parsing routine.

    :param argv: A list of argument strings.
    :rtype argv: list
    :return: A parsed and verified arguments namespace.
    :rtype: :py:class:`argparse.Namespace`
    """
    from argparse import ArgumentParser

    parser = ArgumentParser(
        description=(
            'Network Topology Framework using NML, '
            'with support for pytest.'
        )
    )
    parser.add_argument(
        '-v', '--verbose',
        help='Increase verbosity level',
        default=0,
        action='count'
    )
    parser.add_argument(
        '--version',
        action='version',
        version='Network Topology Framework v{}'.format(__version__)
    )

    parser.add_argument(
        '--platform',
        default=DEFAULT_PLATFORM,
        help='Platform engine to build the topology with',
        choices=platforms()
    )
    parser.add_argument(
        '--non-interactive',
        help='Just build the topology and exit',
        action='store_true'
    )
    parser.add_argument(
        '--show-build-commands',
        help='Show commands executed in nodes during build',
        action='store_true'
    )
    parser.add_argument(
        '--plot-dir',
        default=None,
        help='Directory to auto-plot topologies'
    )
    parser.add_argument(
        '--plot-format',
        default='svg',
        help='Format for plotting topologies'
    )
    parser.add_argument(
        '--nml-dir',
        default=None,
        help='Directory to export topologies as NML XML'
    )
    parser.add_argument(
        '--inject',
        default=None,
        help='Path to an attributes injection file'
    )
    parser.add_argument(
        '--log-dir',
        default=None,
        help='Directory to create log files'
    )

    parser.add_argument(
        'topology',
        help='File with the topology description to build'
    )

    args = parser.parse_args(argv)
    args = validate_args(args)
    return args


__all__ = ['parse_args']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
topology platforms manager module.

This module finds out which topology platforms plugins are installed and
returns them in a dictionary.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from inspect import isclass

from pkg_resources import iter_entry_points

from .platform import BasePlatform


log = logging.getLogger(__name__)


DEFAULT_PLATFORM = 'debug'
"""
Default platform engine.
"""


def platforms(cache=True):
    """
    List all available platform engines.

    This function lists the default engine plus any other it can discover
    looking up the entry point.

    :param bool cache: If ``True`` return the cached result. If ``False`` force
     lookup for plugins registered for the entry point.
    :rtype: list
    :return: A sorted list with all available platforms.
    """

    # Return cached value if call is repeated
    if cache and hasattr(platforms, 'available'):
        return list(platforms.available)

    # Add default plugin
    available = []

    # Iterate over entry points
    for ep in iter_entry_points(group='topology_platform_10'):
        available.append(ep.name)

    available.sort()
    platforms.available = available

    return list(available)


def load_platform(name):
    """
    Load platform identified by given name.

    :param str name: Name of the platform.
     This must be a name available in the list returned by :func:`platforms`.
    :rtype: A :class:`BasePlatform` subclass
    :return: The implementation class on the platform engine.
    """
    if name not in platforms():
        raise RuntimeError('Unknown platform engine "{}".'.format(name))

    # Iterate over entry points
    for ep in iter_entry_points(group='topology_platform_10', name=name):

        try:
            platform = ep.load()
        except Exception as e:
            log.error(
                'Unable to load topology engine '
                'platform plugin {}.'.format(name)
            )
            raise e

        if not isclass(platform) or not issubclass(platform, BasePlatform):
            log.error(
                'Platform "{}" doesn\'t implement the required interface: '
                'Platform not a subclass of BasePlatform.'.format(name)
            )
            continue

        return platform

    raise RuntimeError(
        'Platform engine "{}"" not in entry points.'.format(name)
    )


__all__ = ['platforms', 'load_platform', 'DEFAULT_PLATFORM']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
topology libraries manager module.

This module finds out which topology communication libraries plugins are
installed and returns them in a dictionary.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from functools import partial
from inspect import isfunction, isclass
from traceback import format_exc
from collections import OrderedDict
from argparse import Namespace

from pkg_resources import iter_entry_points


log = logging.getLogger(__name__)


def libraries(cache=True):
    """
    List all available communication libraries.

    This function list all communication libraries it can discover looking up
    the entry point. This lookup requires to load all available communication
    libraries plugins, which can be costly or error prone if a plugin
    misbehave, and because of this a cache is stored after the first call.

    :param bool cache: If ``True`` return the cached result. If ``False`` force
     reload of all communication libraries registered for the entry point.
    :rtype: dict
    :return: A dictionary associating the name of the communication library and
     it functions or classes:

     ::

        {
            'comm1': [function_a, function_b],
            'my_lib': [my_function_a, another_function]
        }
    """

    # Return cached value if call is repeated
    if cache and hasattr(libraries, 'available'):
        return libraries.available

    # Add default plugins
    # Note: None so far
    available = {}

    # Iterate over entry points
    for ep in iter_entry_points(group='topology_library_10'):

        name = ep.name

        try:
            library = ep.load()
        except:
            log.error(
                'Unable to load topology communication '
                'library plugin {}.'.format(name)
            )
            log.debug(format_exc())
            continue

        # Validate library
        if not hasattr(library, '__all__'):
            log.error(
                'Ignoring library "{}" as it doesn\'t '
                'match the required interface: '
                '__all__ is missing.'.format(name)
            )
            continue

        # Validate entities
        invalid = [
            func for func in library.__all__
            if not isfunction(getattr(library, func, None)) and
            not isclass(getattr(library, func, None))
        ]

        if invalid:
            log.error(
                'Ignoring library "{}". '
                'Found non-functions or classes: {}.'.format(
                    name, ', '.join(invalid)
                )
            )
            continue

        available[name] = [
            getattr(library, func, None) for func in library.__all__
        ]

    libraries.available = available
    return available


class LibsProxy(object):
    """
    Proxy object to call communication libraries.

    This proxy object is expected to be used by an engine node to add support
    for communication libraries.

    :param enode: The engine node to communicate with.
    :type enode: topology.platforms.node.BaseNode
    """
    def __init__(self, enode):
        self._enode = enode
        self._libraries = OrderedDict()

        for libname, callables in libraries().items():
            # We create a dictionary using dictionary comprehension syntax
            # that will map for each callable in the library the name of
            # such callable with a new partial function that will bind the
            # first argument to the enode. Then, we expand that dictionary and
            # feed it as kwargs to the Namespace class, that will allow us
            # to use the dictionary keys as instance attributes.
            # Very nice Python magic indeed.
            self._libraries[libname] = Namespace(**{
                c.__name__: partial(c, enode) for c in callables
            })

    def __getattr__(self, name):
        if name not in self._libraries:
            raise Exception(
                'Unknown communication library function {}. '
                'Are you missing a dependency?'.format(name)
            )
        return self._libraries[name]


__all__ = ['LibsProxy', 'libraries']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
topology manager module.

This module takes care of understanding the topology defined in the test
suite, commanding its construction to appropriate topology platform plugin
and finally commanding its destruction to the plugin also.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from sys import exc_info
from copy import deepcopy
from datetime import datetime
from traceback import format_exc
from collections import OrderedDict

from six import string_types

from pynml.manager import ExtendedNMLManager

from .parser import parse_txtmeta
from .platforms.node import BaseNode
from .platforms.manager import platforms, load_platform, DEFAULT_PLATFORM


log = logging.getLogger(__name__)


class TopologyManager(object):
    """
    Main Topology Manager object.

    This object is responsible to build a topology given a description.

    There is three options to build a topology:

    - Using the simplified textual description using a Graphviz like syntax.
      To use this option call the :meth:`parse`.
    - Using a basic Python dictionary to load the description of the topology.
      To use this option call the :meth:`load`.
    - Build a full NML topology using NML objects and relations and register
      all the objects in the namespace using the embedded
      :class:`pynml.manager.ExtendedNMLManager`, for example:

      ::

         mgr = TopologyManager()
         sw1 = mgr.create_node(name='My Switch 1')
         sw2 = mgr.create_node(name='My Switch 2')
         sw1p1 = mgr.create_biport(sw1)
         # ...

      See :class:`pynml.manager.ExtendedNMLManager` for more information of
      this usage.

    :param str engine: Name of the platform engine to build the topology.
     See :func:`platforms` for how to get and discover available platforms.
    """

    def __init__(self, engine=DEFAULT_PLATFORM, **kwargs):

        if engine not in platforms():
            raise RuntimeError('Unknown platform engine "{}".'.format(engine))

        self.nml = ExtendedNMLManager(**kwargs)
        self.engine = engine
        self.nodes = OrderedDict()
        self.ports = OrderedDict()

        self._platform = None
        self._built = False

    def load(self, dictmeta, inject=None):
        """
        Load a topology description in a dictionary format.

        Dictionary format:

        ::

            {
                'nodes': [
                    {
                        'nodes': ['sw1', 'hs1', ...],
                        'attributes': {...}
                    },
                ],
                'ports': [
                    {
                        'ports': [('sw1', '1'), ('hs1', '1'), ...],
                        'attributes': {...}
                    }
                ]
                'links': [
                    {
                        'endpoints': (('sw1', '1'), ('hs1', '1')),
                        'attributes': {...}
                    },
                ]
            }

        See also the module :mod:`topology.parser`.

        :param dict dictmeta: The dictionary to load the topology from.
        :param dict inject: An attributes injection sub-dictionary as defined
         by :func:`parse_attribute_injection`.
        """
        # Load nodes
        for nodes_spec in dictmeta.get('nodes', []):
            for node_id in nodes_spec['nodes']:

                # Explicitly create node
                attrs = deepcopy(nodes_spec['attributes'])
                attrs['identifier'] = node_id

                # Inject the run-specific attributes
                if inject is not None and node_id in inject:
                    for key, value in inject[node_id].items():
                        attrs[key] = value

                self.nml.create_node(**attrs)

        # Load ports
        for ports_spec in dictmeta.get('ports', []):
            for node_id, port in ports_spec['ports']:

                # Auto-create node
                node = self.nml.get_object(node_id)
                if node is None:
                    node = self.nml.create_node(identifier=node_id)

                # Explicitly create port
                attrs = deepcopy(ports_spec['attributes'])
                attrs['identifier'] = '{}-{}'.format(node_id, port)
                attrs['label'] = port
                self.nml.create_biport(node, **attrs)

        # Load links
        for link_spec in dictmeta.get('links', []):

            # Get endpoints
            endpoints = [None, None]
            for idx, (node_id, port) in enumerate(link_spec['endpoints']):

                # Auto-create node
                node = self.nml.get_object(node_id)
                if node is None:
                    node = self.nml.create_node(identifier=node_id)

                # Auto-create biport
                port_id = '{}-{}'.format(node_id, port)
                biport = self.nml.get_object(port_id)
                if biport is None:
                    attrs = {
                        'identifier': port_id,
                        'label': port
                    }
                    biport = self.nml.create_biport(node, **attrs)

                # Register endpoint
                endpoints[idx] = biport

            # Explicit-create links
            attrs = deepcopy(link_spec['attributes'])
            self.nml.create_bilink(*endpoints, **attrs)

    def parse(self, txtmeta, load=True, inject=None):
        """
        Parse a textual topology meta-description.

        For a description of the textual format see the module
        :mod:`topology.parser`.

        :param str txtmeta: The textual meta-description of the topology.
        :param bool load: If ``True`` (the default) call :meth:`load`
         immediately after parse.
        :param dict inject: An attributes injection sub-dictionary as defined
         by :func:`parse_attribute_injection`.
        """
        data = parse_txtmeta(txtmeta)
        if load:
            self.load(data, inject=inject)
        return data

    def is_built(self):
        """
        Check if the current topology was built.

        :rtype: bool
        :return: True if the topology was successfully built.
        """
        return self._built

    def build(self):
        """
        Build the topology hold by this manager.

        This method instance the platform engine and request the build of the
        topology defined.
        """
        if self._built:
            raise RuntimeError(
                'You cannot build a topology twice.'
            )

        node_enode_map = OrderedDict()
        timestamp = datetime.now().replace(microsecond=0).isoformat()
        stage = 'instance'

        # Instance platform
        plugin = load_platform(self.engine)
        self._platform = plugin(timestamp, self.nml)

        try:
            stage = 'pre_build'
            self._platform.pre_build()

            stage = 'add_node'
            for node in self.nml.nodes():
                enode = self._platform.add_node(node)

                # Check that engine node implements the minimum interface
                if not isinstance(enode, BaseNode):
                    msg = (
                        'Platform {} returned an invalid '
                        'engine node {}.'
                    ).format(self.engine, enode)
                    log.critical(msg)
                    raise Exception(msg)

                # Register engine node
                node_enode_map[node.identifier] = enode.identifier
                self.nodes[enode.identifier] = enode
                # Register empty port map
                self.ports[enode.identifier] = OrderedDict()

            stage = 'add_biport'
            for node, biport in self.nml.biports():
                eport = self._platform.add_biport(node, biport)

                # Check that engine port is of correct type
                if not isinstance(eport, string_types):
                    msg = (
                        'Platform {} returned an invalid '
                        'engine port name {}.'
                    ).format(self.engine, enode)
                    log.critical(msg)
                    raise Exception(msg)

                # Register engine port
                label = biport.metadata.get('label', biport.identifier)
                enode_id = node_enode_map[node.identifier]
                self.ports[enode_id][label] = eport

            stage = 'add_bilink'
            for node_porta, node_portb, bilink in self.nml.bilinks():
                self._platform.add_bilink(node_porta, node_portb, bilink)

            stage = 'post_build'

            # Assign the port mapping to the enode so they know their mapping
            # and be able to change it if required
            # (and globally, we do not use deepcopy).
            for enode_id, enode in self.nodes.items():
                enode.ports = self.ports[enode_id]

            self._platform.post_build()

        except:
            e = exc_info()[1]
            log.critical(
                (
                    'Build failed at stage "{}" with "{}". '
                    'Calling plugin rollback routine...'
                ).format(stage, e)
            )
            log.debug(format_exc())
            self._platform.rollback(stage, self.nodes, e)
            raise

        self._built = True

    def unbuild(self):
        """
        Undo the topology.

        This removes all references to the engine nodes and request the
        platform to destroy the topology.
        """
        if not self._built:
            raise RuntimeError(
                'You cannot unbuild and never built topology.'
            )

        # Remove own reference to enodes
        self.nodes = OrderedDict()

        # Call platform destroy hook
        self._platform.destroy()

        # Explicitly delete platform
        del self._platform
        self._platform = None

    def get(self, identifier):
        """
        Get a platform engine with given identifier.

        :param str identifier: The node identifier.
        :rtype: A subclass of :class:`topology.platforms.node.BaseNode`
        :return: The engine node implementing the communication with the node.
        """
        return self.nodes.get(identifier, None)

    def relink(self, link_id):
        """
        Relink back a link specified in the topology.

        :param str link_id: Link identifier to be recreated.
        """
        if not self._built:
            raise RuntimeError(
                'You cannot relink on a never built topology.'
            )
        self._platform.relink(link_id)

    def unlink(self, link_id):
        """
        Unlink (break) a link specified in the topology.

        :param str link_id: Link identifier to be recreated.
        """
        if not self._built:
            raise RuntimeError(
                'You cannot unlink on a never built topology.'
            )
        self._platform.unlink(link_id)


__all__ = ['TopologyManager']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Debug platform engine module for topology.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from datetime import datetime

from .node import CommonNode
from .platform import BasePlatform


log = logging.getLogger(__name__)


class DebugPlatform(BasePlatform):
    """
    Plugin to build a topology for debugging.

    See :class:`topology.platforms.platform.BasePlatform` for more information.
    """

    def __init__(self, timestamp, nmlmanager):
        super(DebugPlatform, self).__init__(timestamp, nmlmanager)

    def pre_build(self):
        """
        See :meth:`BasePlatform.pre_build` for more information.
        """
        log.debug('[HOOK] pre_build')

    def add_node(self, node):
        """
        See :meth:`BasePlatform.add_node` for more information.
        """
        log.debug('[HOOK] add_node({})'.format(
            node
        ))
        return DebugNode(node.identifier, name=node.name, **node.metadata)

    def add_biport(self, node, biport):
        """
        See :meth:`BasePlatform.add_biport` for more information.
        """
        log.debug('[HOOK] add_biport({}, {})'.format(
            node, biport
        ))
        return biport.metadata.get('label', biport.identifier)

    def add_bilink(self, nodeport_a, nodeport_b, bilink):
        """
        See :meth:`BasePlatform.add_bilink` for more information.
        """
        log.debug('[HOOK] add_bilink({}, {}, {})'.format(
            nodeport_a, nodeport_b, bilink
        ))

    def post_build(self):
        """
        See :meth:`BasePlatform.post_build` for more information.
        """
        log.debug('[HOOK] post_build()')

    def destroy(self):
        """
        See :meth:`BasePlatform.destroy` for more information.
        """
        log.debug('[HOOK] destroy()')

    def rollback(self, stage, enodes, exception):
        """
        See :meth:`BasePlatform.rollback` for more information.
        """
        log.debug('[HOOK] rollback({}, {}, {})'.format(
            stage, enodes, exception
        ))

    def relink(self, link_id):
        """
        See :meth:`BasePlatform.relink` for more information.
        """
        log.debug('[CALL] relink({})'.format(
            link_id
        ))

    def unlink(self, link_id):
        """
        See :meth:`BasePlatform.unlink` for more information.
        """
        log.debug('[CALL] unlink({})'.format(
            link_id
        ))


class DebugNode(CommonNode):
    """
    Engine Node for debugging.
    """

    def __init__(self, identifier, **kwargs):
        super(DebugNode, self).__init__(identifier, **kwargs)

    def send_command(self, cmd, shell=None, silent=False):
        """
        Implementation of the ``send_command`` interface.

        This test node will just echo the command.

        See :meth:`CommonNode.send_command` for more information.
        """
        log.debug('{} [{}].send_command(\'{}\', shell=\'{}\') ::'.format(
            datetime.now().isoformat(), str(self), cmd, shell
        ))
        return cmd

    def _get_services_address(self):
        """
        Implementation of the ``_get_services_address`` interface.

        This test node will return '127.0.0.1'.

        See :meth:`CommonNode._get_services_address` for more information.
        """
        return '127.0.0.1'

    def __str__(self):
        return 'DebugNode(identifier={}, metadata={})'.format(
            self.identifier, self.metadata
        )


__all__ = ['DebugPlatform', 'DebugNode']



#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Base platform engine module for topology.

This module defines the functionality that a Topology Engine Node must
implement to be able to create a network using a specific environment.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from datetime import datetime
from collections import OrderedDict
from abc import ABCMeta, abstractmethod

from six import add_metaclass, iterkeys

from .service import BaseService
from ..libraries.manager import LibsProxy
from .shell import ShellContext, BaseShell


log = logging.getLogger(__name__)


@add_metaclass(ABCMeta)
class HighLevelShellAPI(object):
    """
    API used to interact with node shells.

    :var str default_shell: Engine node default shell.
    """

    @property
    def default_shell(self):
        raise NotImplementedError('default_shell')

    @default_shell.setter
    def default_shell(self, value):
        raise NotImplementedError('default_shell.setter')

    @abstractmethod
    def available_shells(self):
        """
        Get the list of available shells.

        :return: The list of all available shells. The first element is the
         default (if any).
        :rtype: List of str.
        """

    @abstractmethod
    def send_command(self, cmd, shell=None, silent=False):
        """
        Send a command to this engine node.

        :param str cmd: Command to send.
        :param str shell: Shell that must interpret the command.
         ``None`` for the default shell. Is up to the engine node to
         determine what its default shell is.
        :param bool silent: True to call the shell logger, False
         otherwise.

        :return: The response of the command.
        :rtype: str
        """

    def __call__(self, *args, **kwargs):
        return self.send_command(*args, **kwargs)

    @abstractmethod
    def _register_shell(self, name, shellobj):
        """
        Allow plugin developers to register a shell when initializing a node.

        :param str name: Unique name of the shell to register.
        :param shellobj: The shell object to register.
        :type shellobj: BaseShell
        """


@add_metaclass(ABCMeta)
class LowLevelShellAPI(object):
    """
    API used to interact with low level shell objects.
    """

    @abstractmethod
    def use_shell(self, shell):
        """
        Create a context manager that allows to use a different default shell
        in a context, including access to it's low-level shell object.

        :param str shell: The default shell to use in the context.

        Assuming, for example, that a node has two shells:
        ``bash`` and ``python``:

        ::

            with mynode.use_shell('python') as python:
                # This context manager sets the default shell to 'python'
                mynode('from os import getcwd')
                cwd = mynode('print(getcwd())')

                # Access to the low-level shell API
                python.send_command('foo = (', matches=['... '])
                ...
        """

    @abstractmethod
    def get_shell(self, shell):
        """
        Get the shell object associated with the given name.

        The shell object allows to access the low-level shell API.

        :param str shell: Name of the shell.

        :return: The associated shell object.
        :rtype: BaseShell
        """


@add_metaclass(ABCMeta)
class ServicesAPI(object):
    """
    API to gather information and connection parameters to a node services.
    """

    @abstractmethod
    def available_services(self):
        """
        Get the list of all available services.

        :return: The list of all available services.
        :rtype: List of str.
        """

    @abstractmethod
    def get_service(self, service):
        """
        Get the service object associated with the given name.

        The service object holds all connection parameters.

        :param str service: Name of the service.

        :return: The associated service object.
        :rtype: BaseService
        """

    @abstractmethod
    def _register_service(self, name, serviceobj):
        """
        Allow plugin developers to register a service when initializing a node.

        :param str name: Unique name of the service to register.
        :param serviceobj: The service object to register with all connection
         related parameters.
        :type serviceobj: BaseService
        """

    @abstractmethod
    def _get_services_address(self):
        """
        Returns the IP or FQDN of the node so users can connect to the
        registered services.

        .. note::

           This method should be implemented by the platform engine base node.

        :returns: The IP or FQDM of the node.
        :rtype: str
        """


@add_metaclass(ABCMeta)
class StateAPI(object):
    """
    API to control the enable/disabled state of a node.
    """

    @abstractmethod
    def is_enabled(self):
        """
        Query if the node is enabled.

        :return: True if the node is enabled, False otherwise.
        :rtype: bool
        """

    @abstractmethod
    def enable(self):
        """
        Enable this node.

        An enabled node allows communication.
        """

    @abstractmethod
    def disable(self):
        """
        Disable this node.

        A disabled node doesn't allow communication.
        """


@add_metaclass(ABCMeta)
class BaseNode(HighLevelShellAPI, LowLevelShellAPI, ServicesAPI, StateAPI):
    """
    Base engine node class.

    This class represent the base interface that engine nodes require to
    implement.

    See the :doc:`Plugins Development Guide </plugins>` for reference.

    :param str identifier: User identifier of the engine node.

    :var identifier: User identifier of the engine node. Please note that
     this identifier is unique in the topology being built, but is not unique
     in the execution system.
    :var metadata: Additional metadata (kwargs leftovers).
    :var ports: Mapping between node ports and engine ports. This variable
     is populated by the :class:`topology.manager.TopologyManager`.
    """

    @abstractmethod
    def __init__(self, identifier, **kwargs):
        self.identifier = identifier
        self.metadata = kwargs
        self.ports = OrderedDict()


@add_metaclass(ABCMeta)
class CommonNode(BaseNode):
    """
    Base engine node class with a common base implementation.

    This class provides a basic common implementation for managing shells and
    services. Internal ordered dictionaries handles the keys for
    shells and services objects that implements the logic for those shells or
    services.

    Child classes will then only require to call registration methods
    :meth:`_register_shell` and :meth:`_register_service`.

    In particular, this class implements support for Communication Libraries
    using class :class:`LibsProxy` that will hook with all available libraries.

    See :class:`BaseNode`.

    .. note::

       The method :meth: ``_get_services_address`` should be provided by the
       Platform Engine base node.
    """

    @abstractmethod
    def __init__(self, identifier, **kwargs):
        super(CommonNode, self).__init__(identifier, **kwargs)

        # Shell(s) API
        self._default_shell = None
        self._shells = OrderedDict()

        # Services API
        self._services = OrderedDict()

        # State API
        self._enabled = True

        # Communication Libraries support
        self.libs = LibsProxy(self)

    # HighLevelShellAPI

    @property
    def default_shell(self):
        return self._default_shell

    @default_shell.setter
    def default_shell(self, value):
        if value not in self._shells:
            raise KeyError(
                'Cannot set default shell. Unknown shell "{}"'.format(value)
            )
        self._default_shell = value

    def available_shells(self):
        """
        Implementation of the public ``available_shells`` interface.

        This method will just list the available keys in the internal ordered
        dictionary.

        See :meth:`HighLevelShellAPI.available_shells` for more information.
        """
        return list(iterkeys(self._shells))

    def send_command(self, cmd, shell=None, silent=False):
        """
        Implementation of the public ``send_command`` interface.

        This method will lookup for the shell argument in an internal ordered
        dictionary to fetch a shell object to delegate the command to.
        If None is provided, the default shell of the node will be used.

        See :meth:`HighLevelShellAPI.send_command` for more information.
        """

        # Check at least one shell is available
        if not self._shells:
            raise Exception(
                'Node {} doens\'t have any shell.'.format(self.identifier)
            )

        # Check if default shell is already set
        if self._default_shell is None:
            self._default_shell = list(iterkeys(self._shells))[0]

        # Check requested shell is supported
        if shell is None:
            shell = self._default_shell
        elif shell not in self._shells.keys():
            raise Exception(
                'Shell {} is not supported.'.format(shell)
            )

        active_shell = self.get_shell(shell)

        active_shell.send_command(cmd, silent=silent)

        response = active_shell.get_response(silent=silent)

        return response

    def _register_shell(self, name, shellobj):
        """
        Implementation of the private ``_register_shell`` interface.

        This method will lookup for the shell name argument in an internal
        ordered dictionary and, if inexistent, it will register the given
        shell object.

        See :meth:`HighLevelShellAPI._register_shell` for more information.
        """
        assert isinstance(shellobj, BaseShell)

        if name in self._shells:
            raise KeyError('Shell "{}" already registered'.format(name))
        if not name:
            raise KeyError('Invalid name for shell "{}"'.format(name))

        self._shells[name] = shellobj

        # Add the node identifier and the shell name to the shell object to
        # enable logging in the shell object itself
        shellobj._register_node(self.identifier, name)

    # LowLevelShellAPI

    def get_shell(self, shell):
        """
        Implementation of the public ``get_shell`` interface.

        This method will return the shell object associated with the given
        shell name.

        See :meth:`LowLevelShellAPI.get_shell` for more information.
        """
        if shell not in self._shells:
            raise KeyError(
                'Unknown shell "{}"'.format(shell)
            )
        return self._shells[shell]

    def use_shell(self, shell):
        """
        Implementation of the public ``use_shell`` interface.

        This method allows to create contexts (using a Python Context Manager)
        that allows the user, by the means of a ``with`` statement, to create
        a context to use a specific shell in it.

        See :meth:`LowLevelShellAPI.use_shell` for more information.
        """
        return ShellContext(self, shell)

    # ServicesAPI

    def available_services(self):
        """
        Implementation of the public ``available_services`` interface.

        This method will just list the available keys in the internal ordered
        dictionary.

        See :meth:`ServicesAPI.available_services` for more information.
        """
        return list(iterkeys(self._services))

    def get_service(self, service):
        """
        Implementation of the public ``get_service`` interface.

        This method will return the service object associated with the given
        service name.

        See :meth:`ServicesAPI.get_service` for more information.
        """
        if service not in self._services:
            raise KeyError(
                'Unknown service "{}"'.format(service)
            )

        # Set the node address
        serviceobj = self._services[service]
        serviceobj.address = self._get_services_address()

        return serviceobj

    def _register_service(self, name, serviceobj):
        """
        Implementation of the private ``_register_service`` interface.

        This method will lookup for the service name argument in an internal
        ordered dictionary and, if inexistent, it will register the given
        service object.

        See :meth:`ServicesAPI._register_service` for more information.
        """
        assert isinstance(serviceobj, BaseService)

        if name in self._services:
            raise KeyError('Service "{}" already registered'.format(name))
        if not name:
            raise KeyError('Invalid name for service "{}"'.format(name))

        self._services[name] = serviceobj

    # StateAPI

    def is_enabled(self):
        """
        Implementation of the ``is_enabled`` interface.

        This method will just return the internal value of the ``_enabled``
        flag.

        See :meth:`StateAPI.is_enabled` for more information.
        """
        return self._enabled

    def enable(self):
        """
        Implementation of the ``enable`` interface.

        This method will just set the value of the ``_enabled`` flag to True.

        See :meth:`StateAPI.enable` for more information.
        """
        self._enabled = True

    def disable(self):
        """
        Implementation of the ``disable`` interface.

        This method will just set the value of the ``_enabled`` flag to False.

        See :meth:`StateAPI.disable` for more information.
        """
        self._enabled = False

    def _log_command(self, command, shell):
        """
        Command logging function for low-level shell API usage.

        :param str command: Sent command to be logged.
        :param str shell: Name of the shell that sends the command.
        """

        print(
            '{} [{}].send_command(\'{}\', shell=\'{}\') ::'.format(
                datetime.now().isoformat(), self.identifier, command, shell
            )
        )

    def _log_response(self, response, shell):
        """
        Response logging function for low-level shell API usage.

        :param str response: Command response to be logged.
        :param str shell: Name of the shell that receives the command response.
        """
        print(response.encode(self._shells[shell]._encoding))

__all__ = [
    'HighLevelShellAPI',
    'LowLevelShellAPI',
    'ServicesAPI',
    'StateAPI',
    'BaseNode',
    'CommonNode'
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Logging module for the Topology Modular Framework.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from abc import ABCMeta
from os.path import join
from inspect import stack
from collections import OrderedDict
from distutils.dir_util import mkpath
from weakref import WeakValueDictionary

from six import add_metaclass


LEVELS = OrderedDict([
    ('NOTSET', logging.NOTSET),
    ('DEBUG', logging.DEBUG),
    ('INFO', logging.INFO),
    ('WARNING', logging.WARNING),
    ('ERROR', logging.ERROR),
    ('CRITICAL', logging.CRITICAL),
])
"""Simple map with the default logging levels."""


@add_metaclass(ABCMeta)
class BaseLogger(object):
    """
    Base class for Topology logger classes.

    :param OrderedDict nameparts: Namespaced name of the logger.
    :param bool propagate: Propagate messages to the root logger.
    :param int level: Logging level as in Python's logging framework.
     .. autodata:: topology.logging.LEVELS
    :param str log_dir: Directory to store the logging file
     (if any, see subclasses).

    Object read-only properties:

    :var OrderedDict nameparts: Namespaced name of the logger.
     For example::

         OrderedDict([
             'context': 'test_bar',
             'category': 'core',
             'name': 'mylogger'
         ])

    :var str name: Name of this logger as concatenated nameparts.
     For example:

         test_bar.core.mylogger

    Object read-write properties:

    :var bool propagate: True if message propagation to the root logger is
     enabled.
    :var int level: Current logging level for this logger.
    :var log_dir: Current logging directory for this logger.
    """

    def __init__(
        self, nameparts,
        propagate=False,
        level=logging.NOTSET,
        log_dir=None,
        *args, **kwargs
    ):
        self._nameparts = nameparts
        self._propagate = None
        self._level = None
        self._log_dir = None

        self._name = ''.join(map(str, nameparts.values()))
        self.logger = logging.getLogger(self._name)

        self.propagate = propagate
        self.level = level
        self.log_dir = log_dir

    @property
    def nameparts(self):
        return OrderedDict(self._nameparts)

    @property
    def name(self):
        return self._name

    @property
    def propagate(self):
        return self._propagate

    @propagate.setter
    def propagate(self, propagate):
        self.logger.propagate = propagate

    @property
    def level(self):
        return self._level

    @level.setter
    def level(self, level):
        self._level = level
        self.logger.setLevel(self._level)

    @property
    def log_dir(self):
        return self._log_dir

    @log_dir.setter
    def log_dir(self, log_dir):
        self._log_dir = log_dir


@add_metaclass(ABCMeta)
class StdOutLogger(BaseLogger):
    """
    Logger that logs to the standard output.
    """
    def __init__(self, *args, **kwargs):
        super(StdOutLogger, self).__init__(*args, **kwargs)

        self.logger.addHandler(logging.StreamHandler())

    def log(self, message):
        """
        Log a message to standard output.

        :param str message: Message to log.
        """
        self.logger.log(self._level, message)


@add_metaclass(ABCMeta)
class FileLogger(BaseLogger):
    """
    Subclass of BaseLogger that adds a FileHandler.

    This class implements additional logic for when the ``log_dir`` path is
    changed. It will create a file with the name of the logger under the
    ``log_dir`` with the ``.log`` extension.

    :param str file_formatter: Format to use in the FileHandler, defaulting to
     ``logging.BASIC_FORMAT``.
    """
    def __init__(self, *args, **kwargs):
        self._file_handler = None
        self._file_formatter = kwargs.pop(
            'file_formatter', logging.BASIC_FORMAT
        )
        super(FileLogger, self).__init__(*args, **kwargs)

    @BaseLogger.log_dir.setter
    def log_dir(self, log_dir):

        # Do not recreate a new file handler if nothing has changed
        if log_dir == self._log_dir:
            return

        def reset_file_handler():
            self.logger.removeHandler(self._file_handler)
            self._file_handler.close()
            self._file_handler = None

        if log_dir:

            # Remove old file handler
            if self._file_handler:
                reset_file_handler()

            # Create file handler
            fh = logging.FileHandler(
                join(log_dir, '{}.{}'.format(self._name, 'log'))
            )

            # Create formatter
            formatter = logging.Formatter(self._file_formatter)
            fh.setFormatter(formatter)

            # Register handler
            self.logger.addHandler(fh)
            self._file_handler = fh

        elif self._file_handler:
            reset_file_handler()

        self._log_dir = log_dir


class PexpectLogger(FileLogger):
    """
    Special subclass that implements a logger to be used with Pexpect
    ``logfile`` keyword argument.

    **Purpose:** to log every character in the pexpect session.

    Pexpect ``logfile`` is expected to be an open file-like object, so this
    class implements the ``write()`` and ``flush()`` operations for effective
    duck-typing.

    To implement one logging per ``flush()`` this class implements a local
    buffer.

    Pexpect loggers can be located with the name::

        <context>.pexpect.<node_identifier>.<shell_name>.<connection>

    Where ``<context>`` is optional if running interactively or if the context
    in the LoggerManager has not been set.

    :param str encoding: Encoding to log into. The ``write()`` interface expect
     bytes, so in order to log in plain text it is required to perform a
     decoding. Defaults to ``utf-8``.
    :param str errors: Handling of decoding errors. The ``write()`` interface
     may find undecodeable characters when decoding with the selected encoding,
     so the handling of errors needs to be defined. The values available are
     the same ones that the ``decode`` method of a bytes object expects in its
     ``error`` keyword argument. Defaults to ``ignore``.
    """
    def __init__(self, *args, **kwargs):
        self._encoding = kwargs.pop('encoding', 'utf-8')
        self._errors = kwargs.pop('errors', 'ignore')
        self._buffer = []

        if 'file_formatter' not in kwargs:
            kwargs['file_formatter'] = '%(message)s'

        super(PexpectLogger, self).__init__(*args, **kwargs)

    def write(self, data):
        self._buffer.append(
            data.decode(encoding=self._encoding, errors=self._errors)
        )

    def flush(self):
        data = ''.join(self._buffer)
        del self._buffer[:]
        self.logger.log(self._level, data)


class ConnectionLogger(BaseLogger):
    """
    Logger for shell connections.

    **Purpose:** to log every ``send_command`` and ``get_response`` in a
    low-level connection.

    Connection loggers can be located with the name::

        <context>.pexpect.<node_identifier>.<shell_name>.<connection>

    Where ``<context>`` is optional if running interactively or if the context
    in the LoggerManager has not been set.
    """
    def __init__(self, *args, **kwargs):
        super(ConnectionLogger, self).__init__(*args, **kwargs)
        self.logger.addHandler(logging.StreamHandler())

    def log_send_command(self, command, matches, newline, timeout):
        fields = {
            'command': command,
            'matches': matches,
            'newline': newline,
            'timeout': timeout
        }
        fields.update(self._nameparts)

        log_entry = (
            '[{node_identifier}].get_shell({shell_name}).send_command('
            '\'{command}\', '
            'matches={matches}, newline={newline}, timeout={timeout}'
            ')'
        ).format(**fields)
        self.logger.info(log_entry)

    def log_get_response(self, response):
        fields = {
            'response': response
        }
        fields.update(self._nameparts)

        log_entry = (
            '[{node_identifier}].get_shell({shell_name}).get_response() '
            '->\n'
            '{response}'
        ).format(**fields)
        self.logger.info(log_entry)


class StepLogger(StdOutLogger):
    """
    Stepper logging class.

    This class will log a message and will show the step number and the caller
    name and line number.
    """
    def __init__(
        self, nameparts=OrderedDict(
            [('test_suite', ''), ('test_case', '')]
        ), *args, **kwargs
    ):

        self.step = 0

        for part in ['test_suite', 'test_case']:
            if part not in nameparts:
                raise RuntimeError(
                    'Badly named StepLogger. Missing the {} '
                    'name part: {}'.format(part, nameparts)
                )
            setattr(self, '_{}'.format(part), nameparts[part])

        super(StepLogger, self).__init__(nameparts, *args, **kwargs)

    def __call__(self, msg):
        # Update step count
        self.step += 1

        # Fetch information of the caller
        (
            frame, filename, line_number,
            function_name, lines, index
        ) = stack()[1]

        # Determine execution context
        execution_context = self._test_case
        if function_name != self._test_case:
            execution_context = '{}->{}'.format(
                execution_context,
                function_name
            )
        execution_context = '{}.{}'.format(
            self._test_suite, execution_context
        )

        # Log message
        self.log(
            '>>> [{:03d}] :: {}:{}\n{}'.format(
                self.step, execution_context, line_number,
                '\n'.join(
                    ['... {}'.format(l) for l in msg.strip().splitlines()]
                )
            )
        )


class LoggingManager(object):
    """
    Defines an object that manage and create loggers for Topology components.

    Only an instance of this class is to be created, this instance exists in
    this module and is to be used by other components by importing its
    ``get_logger`` method straigth from this module.

    This method will return a logger tailored for each of these categories,
    if those loggers are implemented:

    #. core
    #. library
    #. platform
    #. user
    #. node
    #. shell
    #. connection
    #. pexpect
    #. service
    #. step

    Read-only properties:

    :var categories: A list of available logging categories.

    Read-write properties:

    :var str logging_directory: Framework wide logging directory. Any change on
     this setting will be notified to all loggers.
    :var str logging_context: Current framework wide logging context.
    """
    def __init__(self, default_level=logging.INFO, default_propagate=False):
        """
        Framework logging manager.
        """
        self._log_dir = None
        self._log_context = None

        self._categories = OrderedDict([
            # module
            ('core', None),
            # name
            ('library', None),
            # name
            ('platform', None),
            # test case
            ('user', None),
            # node identifier
            ('node', None),
            # node identifier, shell name
            ('shell', None),
            # node identifier, shell name, connection
            ('connection', ConnectionLogger),
            # node identifier, shell name, connection
            ('pexpect', PexpectLogger),
            # node identifier, service name, port
            ('service', None),
            # step
            ('step', StepLogger)
        ])
        self._loggers = {
            key: WeakValueDictionary() for key in self._categories.keys()
        }

        self._default_level = default_level
        self._levels = {
            key: self._default_level for key in self._categories.keys()
        }

        self._default_propagate = default_propagate
        self._propagate = {
            key: self._default_propagate for key in self._categories.keys()
        }

    @property
    def categories(self):
        return list(self._categories.keys())

    @property
    def logging_directory(self):
        return self._log_dir

    @logging_directory.setter
    def logging_directory(self, log_dir):
        mkpath(log_dir)
        self._log_dir = log_dir

        # Notify all categories of a log directory change
        for category in self._loggers.values():
            for logger in category.values():
                logger.log_dir = log_dir

    @property
    def logging_context(self):
        return self._log_context

    @logging_context.setter
    def logging_context(self, log_context):
        self._log_context = log_context

        # XXX: The logging context is built-in in the logger name.
        #      By design (?), once a logger was created for a given context,
        #      it cannot change.

    def set_category_level(self, category, level):
        """
        Set the logging level property to all logger in the given category.

        :param str category: Name of the category.
        :param int level: Value to set the level property.
        """
        if category not in self._categories.keys():
            raise ValueError(
                'Unknown category "{}"'.format(category)
            )
        self._levels[category] = level
        for logger in self._loggers[category].values():
            logger.level = level

    def set_category_propagate(self, category, propagate):
        """
        Set the propagate property to all logger in the given category.

        :param str category: Name of the category.
        :param bool propagate: Value of set the propagate property.
        """
        if category not in self._categories.keys():
            raise ValueError(
                'Unknown category "{}"'.format(category)
            )
        self._propagate[category] = propagate
        for logger in self._loggers[category].values():
            logger.propagate = propagate

    def get_logger(self, name, category='core'):
        """
        Get a logger tailored for the given category.

        :param name: Full namespaced name or simple name.
        :type name: OrderedDict or str
        :param str category: Category of the logger.

        :return: A new logger instance of the given category.
        :rtype: BaseLogger
        """
        if category not in self._categories.keys():
            raise ValueError(
                'Unknown category "{}" for logger {}'.format(category, name)
            )

        # Prepend context and category to the nameparts namespace specifier
        nameparts = OrderedDict()
        if self._log_context is not None:
            nameparts['log_context'] = self._log_context
        nameparts['category'] = category

        if isinstance(name, OrderedDict):
            nameparts.update(name)
        else:
            nameparts['name'] = name

        clss = self._categories[category]
        if clss is not None:

            # Instance logger
            instance = clss(
                nameparts,
                propagate=self._propagate[category],
                level=self._levels[category],
                log_dir=self._log_dir
            )

            # Append logger to category registry
            self._loggers[category][id(instance)] = instance
            return instance

        raise NotImplementedError(
            'Category "{}" not implemented'.format(category)
        )


manager = LoggingManager()
"""
Main framework wide instance of :class:LoggingManager.
"""

get_logger = manager.get_logger


__all__ = [
    'manager',
    'get_logger'
]

__api__ = [
    'LEVELS',
    'BaseLogger',
    'FileLogger',
    'PexpectLogger',
    'ConnectionLogger',
    'LoggingManager'
] + __all__



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
topology parser module.

This module takes care of parsing a topology meta-description in a Graphviz
like format.

This topology representation format allows to quickly specify simple topologies
that are only composed of simple nodes, ports and links between them. For a
more programmatic format consider the ``metadict`` format or using the
:class:`pynml.manager.ExtendedNMLManager` directly.

The format for the textual description of a topology is similar to Graphviz
syntax and allows to define nodes and ports with shared attributes and links
between two endpoints with shared attributes too.

::

    # Nodes
    [type=switch attr1=1] sw1 sw2
    hs1

    # Ports
    [speed=1000] sw1:3 sw2:3

    # Links
    [linkattr1=20] sw1:a -- sw2:a
    [linkattr2=40] sw1:3 -- sw2:3

In the above example two nodes with the attributes ``type`` and ``attr1`` are
specified. Then a third node `hs1` with no particular attributes is defined.
Later, we specify some attributes (speed) for a couple of ports. In the same
way, a link between endpoints MAY have attributes. An endpoint is a combination
of a node and a port name.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from traceback import format_exc
from collections import OrderedDict


log = logging.getLogger(__name__)


class ParseException(Exception):
    """
    Custom exception thrown by the parser if it could succeded.

    :var lineno: Line number of the parsing error.
    :var raw_line: Raw line that failed to be parsed.
    :var exc: Internal failure traceback of the parser.
    """
    def __init__(self, lineno, raw_line, exc):
        super(ParseException, self).__init__(
            'Unable to parse line #{}: "{}"'.format(
                lineno, raw_line
            )
        )
        self.lineno = lineno
        self.raw_line = raw_line
        self.exc = exc


def build_parser():
    """
    Build a pyparsing parser for our custom topology description language.

    :rtype: pyparsing.MatchFirst
    :return: A pyparsing parser.
    """
    from pyparsing import (
        Word, Literal, QuotedString,
        StringStart, StringEnd,
        alphas, nums, alphanums,
        Group, OneOrMore, Optional
    )

    number = Word(nums)
    text = QuotedString('"')
    identifier = Word(alphas, alphanums + '_')

    attribute = (
        identifier('key') + Literal('=') +
        (text | number | identifier)('value')
    )
    attributes = (
        Literal('[') +
        OneOrMore(Group(attribute))('attributes') +
        Literal(']')
    )

    node = identifier('node')
    port = node + Literal(':') + (identifier | number)('port')
    link = port('endpoint_a') + Literal('--') + port('endpoint_b')

    nodes_spec = (
        StringStart() + Optional(attributes) +
        OneOrMore(Group(node))('nodes') + StringEnd()
    )
    ports_spec = (
        StringStart() + Optional(attributes) +
        OneOrMore(Group(port))('ports') + StringEnd()
    )
    link_spec = (
        StringStart() + Optional(attributes) +
        link('link') + StringEnd()
    )

    statement = link_spec | ports_spec | nodes_spec

    return statement


def parse_txtmeta(txtmeta):
    """
    Parse a textual description of a topology and return a dictionary of it's
    elements.

    :param str txtmeta: The textual meta-description of the topology.
    :rtype: dict
    :return: Topology as dictionary.
    """
    statement = build_parser()
    data = {
        'nodes': [],
        'ports': [],
        'links': []
    }

    def process_attributes(parsed):
        """
        Convert a pyparsing object with parsed attributes in a dictionary.
        """
        attrs = OrderedDict()

        if 'attributes' not in parsed:
            return attrs

        for attr in parsed.attributes:

            value = attr.value

            # Try to convert simple types
            try:
                value = int(value)
            except ValueError:
                pass
            if value == 'True':
                value = True
            elif value == 'False':
                value = False

            attrs[attr.key] = value
        return attrs

    for lineno, raw_line in enumerate(txtmeta.splitlines(), 1):
        try:
            # Ignore comments and empty line
            line = raw_line.strip()
            if not line or line.startswith('#'):
                continue

            # Parse line and optional attributes
            parsed = statement.parseString(line)
            attrs = process_attributes(parsed)
            log.debug(parsed.dump())
            log.debug(attrs)

            # Process link lines
            if 'link' in parsed:
                link = parsed.link
                data['links'].append({
                    'endpoints': (
                        (link.endpoint_a.node, link.endpoint_a.port),
                        (link.endpoint_b.node, link.endpoint_b.port),
                    ),
                    'attributes': attrs
                })
                continue

            # Process port lines
            if 'ports' in parsed:
                data['ports'].append({
                    'ports': [
                        (port.node, port.port) for port in parsed.ports
                    ],
                    'attributes': attrs
                })
                continue

            # Process port lines
            if 'nodes' in parsed:
                data['nodes'].append({
                    'nodes': [node.node for node in parsed.nodes],
                    'attributes': attrs
                })
                continue

            raise Exception('Unknown line type parsed.')

        except Exception:
            e = ParseException(lineno, raw_line, format_exc())
            log.error(str(e))
            log.debug(e.exc)
            raise e

    return data


def find_topology_in_python(filename):
    """
    Find the TOPOLOGY variable inside a Python file.

    This helper functions build a AST tree a grabs the variable from it. Thus,
    the Python code isn't executed.

    :param str filename: Path to file to search for TOPOLOGY.
    :rtype: str or None
    :return: The value of the TOPOLOGY variable if found, None otherwise.
    """
    import ast

    try:
        with open(filename) as fd:
            tree = ast.parse(fd.read())

        for node in ast.iter_child_nodes(tree):
            if not isinstance(node, ast.Assign):
                continue
            if node.targets[0].id == 'TOPOLOGY':
                return node.value.s

    except:
        log.error(format_exc())
    return None


__all__ = [
    'ParseException',
    'parse_txtmeta',
    'find_topology_in_python'
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from setuptools import setup, find_packages


def read(filename):
    """
    Read a file relative to setup.py location.
    """
    import os
    here = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(here, filename)) as fd:
        return fd.read()


def find_version(filename):
    """
    Find package version in file.
    """
    import re
    content = read(filename)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]", content, re.M
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


def find_requirements(filename):
    """
    Find requirements in file.
    """
    import string
    content = read(filename)
    requirements = []
    for line in content.splitlines():
        line = line.strip()
        if line and line[:1] in string.ascii_letters:
            requirements.append(line)
    return requirements


setup(
    name='topology',
    version=find_version('lib/topology/__init__.py'),
    package_dir={'': 'lib'},
    packages=find_packages('lib'),
    scripts=['bin/topology'],

    # Dependencies
    install_requires=find_requirements('requirements.txt'),

    # Metadata
    author='Hewlett Packard Enterprise Development LP',
    author_email='hpe-networking@lists.hp.com',
    description=(
        'Network Topology Framework using NML, with support for pytest.'
    ),
    long_description=read('README.rst'),
    url='http://topology.rtfd.org/',
    keywords='topology',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
    ],

    # Entry points
    entry_points={
        'pytest11': ['topology = topology.pytest.plugin'],
        'topology_platform_10': [
            'debug = topology.platforms.debug:DebugPlatform'
        ],
        'topology_library_10': [
            'common = topology.libraries.common'
        ]
    }
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
topology pytest plugin module entry point.

This plugin provides a fixture ``topology`` that will load and build a topology
description from the module. This topology must be present in the module as a
constant variable ``TOPOLOGY``. It can be either a string and thus the method
:meth:`topology.manager.TopologyManager.parse` will be used, or a dictionary in
which case the method :meth:`topology.manager.TopologyManager.load` will be
used. Once built, the plugin registers the *unbuild* for when the module has
ended all the tests.

If the ``TOPOLOGY`` variable isn't present the fixture assumes the user will
prefer to build the topology using the standard NML objects with the
:class:`pynml.manager.NMLManager` instance enbeed into the
:class:`topology.manager.TopologyManager`.

To be able to select the platform engine this plugins registers the
``--topology-platform`` option that can be set in pytest command line.

For reference see:

    http://pytest.org/dev/plugins.html#hook-specification-and-validation
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

from os import getcwd, makedirs
from traceback import format_exc
from collections import OrderedDict
from os.path import join, isabs, abspath, exists, isdir

from pytest import fixture, fail, hookimpl, skip

from topology.logging import get_logger, StepLogger


class TopologyPlugin(object):
    """
    pytest plugin for Topology.

    :param str platform: Platform engine name to run the tests with.
    :param str plot_dir: Directory to auto-plot topologies. ``None`` if
     feature is disabled.
    :param str plot_format: Format to plot the topologies.
    :param str nml_dir: Directory to auto-export topologies. ``None`` if
     feature is disabled.
    """

    def __init__(
            self, platform, plot_dir, plot_format,
            nml_dir, injected_attr, log_dir):
        self.platform = platform
        self.plot_dir = plot_dir
        self.plot_format = plot_format
        self.nml_dir = nml_dir
        self.injected_attr = injected_attr
        self.log_dir = log_dir

    def pytest_report_header(self, config):
        """
        pytest hook to print information of the report header.
        """
        header = ["topology: platform='{}'".format(self.platform)]
        if self.plot_dir:
            header.append("          plot_dir='{}' ({})".format(
                self.plot_dir, self.plot_format
            ))
        if self.nml_dir:
            header.append("          nml_dir='{}'".format(
                self.nml_dir
            ))
        if self.log_dir:
            header.append("          log_dir='{}'".format(
                self.log_dir
            ))

        return '\n'.join(header)


@fixture(scope='module')
def topology(request):
    """
    Fixture that injects a TopologyManager into as a test fixture.

    See:

    - https://pytest.org/latest/fixture.html
    - https://pytest.org/latest/builtin.html#_pytest.python.FixtureRequest
    """
    from ..manager import TopologyManager
    from ..logging import manager as logmanager

    plugin = request.config._topology_plugin
    module = request.module
    topomgr = TopologyManager(plugin.platform)

    # Setup framework logging
    logmanager.logging_context = module.__name__
    if plugin.log_dir:
        logmanager.logging_directory = plugin.log_dir

    # Finalizer unbuild the topology and plot it
    def finalizer():

        # Do nothing is topology isn't built
        if not topomgr.is_built():
            return

        # Plot topology
        if plugin.plot_dir:
            plot_file = join(
                plugin.plot_dir,
                '{}.{}'.format(module.__name__, plugin.plot_format)
            )
            topomgr.nml.save_graphviz(
                plot_file, keep_gv=True
            )

        # Export topology as NML
        if plugin.nml_dir:
            nml_file = join(
                plugin.nml_dir,
                '{}.xml'.format(module.__name__)
            )
            topomgr.nml.save_nml(
                nml_file, pretty=True
            )

        topomgr.unbuild()

    # Autobuild topology if available.
    if hasattr(module, 'TOPOLOGY'):

        # Get topology description
        topo = module.TOPOLOGY

        # Get attributes to inject
        suite_injected_attr = None
        if plugin.injected_attr is not None:
            suite_injected_attr = plugin.injected_attr.get(
                abspath(module.__file__), None
            )

        try:
            if isinstance(topo, dict):
                topomgr.load(topo, inject=suite_injected_attr)
            else:
                topomgr.parse(topo, inject=suite_injected_attr)
            topomgr.build()
        except:
            fail(
                'Error building topology in module {}:\n{}'.format(
                    module.__name__,
                    format_exc()
                ), pytrace=False
            )

        request.addfinalizer(finalizer)

    return topomgr


@fixture(scope='function')
def step(request):
    """
    Fixture to log a step in a test.
    """
    return get_logger(
        OrderedDict([
            ('test_suite', request.module.__name__),
            ('test_case', request.function.__name__)
        ]),
        category='step'
    )


def pytest_addoption(parser):
    """
    pytest hook to add CLI arguments.
    """
    from ..platforms.manager import platforms, DEFAULT_PLATFORM

    group = parser.getgroup('topology', 'Testing of network topologies')
    group.addoption(
        '--topology-platform',
        default=DEFAULT_PLATFORM,
        help='Select platform to run topology tests',
        choices=platforms()
    )
    group.addoption(
        '--topology-plot-dir',
        default=None,
        help='Directory to auto-plot topologies'
    )
    group.addoption(
        '--topology-plot-format',
        default='svg',
        help='Format for plotting topologies'
    )
    group.addoption(
        '--topology-nml-dir',
        default=None,
        help='Directory to export topologies as NML XML'
    )
    group.addoption(
        '--topology-inject',
        default=None,
        help='Path to an attributes injection file'
    )
    group.addoption(
        '--topology-log-dir',
        default=None,
        help='Path to a directory where logs are to be stored'
    )


def pytest_configure(config):
    """
    pytest hook to configure plugin.
    """
    # Get registered options
    platform = config.getoption('--topology-platform')
    plot_format = config.getoption('--topology-plot-format')
    plot_dir = config.getoption('--topology-plot-dir')
    nml_dir = config.getoption('--topology-nml-dir')
    injection_file = config.getoption('--topology-inject')
    log_dir = config.getoption('--topology-log-dir')

    def create_dir(path):
        if path:
            if not isabs(path):
                path = join(abspath(getcwd()), path)
            if not exists(path):
                makedirs(path)

    # Determine plot, NML and log directory paths and create them if required
    create_dir(plot_dir)
    create_dir(nml_dir)
    create_dir(log_dir)

    # Parse attributes injection file
    from ..injection import parse_attribute_injection
    injected_attr = None
    if injection_file is not None:

        # Get a list of all testing directories
        search_paths = [
            abspath(arg) for arg in config.args if isdir(arg)
        ]

        injected_attr = parse_attribute_injection(
            injection_file,
            search_paths=search_paths
        )

    # Create and register plugin
    config._topology_plugin = TopologyPlugin(
        platform, plot_dir, plot_format.lstrip('.'),
        nml_dir, injected_attr, log_dir
    )
    config.pluginmanager.register(config._topology_plugin)

    # Add test_id marker
    config.addinivalue_line(
        'markers',
        'test_id(id): assign a test identifier to the test'
    )

    # Add topology_compatible marker
    config.addinivalue_line(
        'markers',
        'platform_incompatible(platforms): '
        'mark a test as incompatible with a list of platform engines'
    )


def pytest_unconfigure(config):
    """
    pytest hook to unconfigure plugin.
    """
    plugin = getattr(config, '_topology_plugin', None)
    if plugin:
        del config._topology_plugin
        config.pluginmanager.unregister(plugin)


@hookimpl(tryfirst=True)
def pytest_runtest_setup(item):
    """
    pytest hook to setup test before run.
    """
    test_id_marker = item.get_marker('test_id')
    incompatible_marker = item.get_marker('platform_incompatible')

    # If marked and xml logging enabled
    if test_id_marker is not None and hasattr(item.config, '_xml'):
        test_id = test_id_marker.args[0]
        item.config._xml.node_reporter(item.nodeid).add_property(
            'test_id', test_id
        )

    if incompatible_marker:
        platform = item.config._topology_plugin.platform
        if platform in incompatible_marker.args[0]:
            skip('Test is incompatible with {} platform'.format(platform))


__all__ = [
    'TopologyPlugin',
    'topology',
    'pytest_addoption',
    'pytest_configure',
    'StepLogger'
]
